(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Instantiation
imports
  Icap 
  QDP_Framework
  Innermost_Usable_Rules
begin


lemma ichain_instantiation: fixes s t \<sigma> f
  defines Pr: "Pr \<equiv> \<lambda> P' i s' t' \<sigma>'. s i \<cdot> \<sigma> i = s' \<cdot> \<sigma>' \<and> t i \<cdot> \<sigma> i = t' \<cdot> \<sigma>' \<and> (s',t') \<in> f P'"
  assumes ichain: "min_ichain (P,Pw,Q,R,Rw) s t \<sigma>"
  and main: "\<And> i P'. i \<ge> n \<Longrightarrow> (s i, t i) \<in> P' \<Longrightarrow> \<exists> s' t' \<sigma>'. Pr P' i s' t' \<sigma>'"
  shows "\<exists> s' t' \<sigma>'. min_ichain (f P, f Pw, Q, R, Rw) s' t' \<sigma>'"
proof -
  let ?QR = "(qrstep Q (R \<union> Rw))^*"
  note ichain = ichain[unfolded min_ichain.simps ichain.simps minimal_cond_def]
  let ?s = "fst"
  let ?t = "fst o snd"
  let ?\<sigma> = "snd o snd"
  let ?Q = "\<lambda> P i st\<sigma>. Pr P i (?s st\<sigma>) (?t st\<sigma>) (?\<sigma> st\<sigma>)"
  let ?st\<sigma> = "\<lambda> i. if (s i, t i) \<in> P 
    then (SOME st\<sigma>. ?Q P i st\<sigma>)
    else (SOME st\<sigma>. ?Q Pw i st\<sigma>)"
  obtain st\<sigma> where st\<sigma>: "st\<sigma> = (\<lambda> i. ?st\<sigma> (i + n))" by auto
  let ?fP = "f P"
  let ?fPw = "f Pw"
  obtain sr where sr: "sr = ?s o st\<sigma>" by auto
  obtain tr where tr: "tr = ?t o st\<sigma>" by auto
  obtain \<sigma>r where \<sigma>r: "\<sigma>r = ?\<sigma> o st\<sigma>" by auto
  {
    fix i
    assume mem: "(s (i + n), t (i + n)) \<in> P"
    from main[OF _ mem] 
    obtain s' t' \<sigma>' where P: "Pr P (i + n) s' t' \<sigma>'" by auto
    let ?st\<sigma>' = "(s',t',\<sigma>')" 
    from P have Q: "?Q P (i + n) ?st\<sigma>'" by simp
    from someI[of "?Q P (i + n)" ?st\<sigma>', OF Q]
    have "?Q P (i + n) (st\<sigma> i)" unfolding st\<sigma> using mem by simp
  } note P = this
  {
    fix i
    assume nomem: "(s (i + n), t (i + n)) \<notin> P"
    with ichain have mem: "(s (i + n), t (i + n)) \<in> Pw" by auto
    from main[OF _ mem] 
    obtain s' t' \<sigma>' where P: "Pr Pw (i + n) s' t' \<sigma>'" by auto
    let ?st\<sigma>' = "(s',t',\<sigma>')" 
    from P have Q: "?Q Pw (i + n) ?st\<sigma>'" by simp
    from someI[of "?Q Pw (i + n)" ?st\<sigma>', OF Q]
    have "?Q Pw (i + n) (st\<sigma> i)" unfolding st\<sigma> using nomem by simp
  } note Pw = this
  {
    fix i
    have ids: "sr i \<cdot> \<sigma>r i = s (i + n) \<cdot> \<sigma> (i + n)"
      using P[of i] Pw[of i] unfolding sr tr \<sigma>r Pr by (cases "(s (i + n), t (i + n)) \<in> P", auto)
    have idt: "tr i \<cdot> \<sigma>r i = t (i + n) \<cdot> \<sigma> (i + n)"
      using P[of i] Pw[of i] unfolding sr tr \<sigma>r Pr by (cases "(s (i + n), t (i + n)) \<in> P", auto)
    have mem: "(sr i, tr i) \<in> ?fP \<union> ?fPw"
      using P[of i] Pw[of i] unfolding sr tr \<sigma>r Pr by (cases "(s (i + n), t (i + n)) \<in> P", auto)
    have strict: "(s (i + n), t (i + n)) \<in> P \<Longrightarrow> (sr i, tr i) \<in> ?fP"
      using P[of i] unfolding sr tr Pr by auto
    note ids idt mem strict
  } note main = this
  have "min_ichain (?fP,?fPw,Q,R,Rw) sr tr \<sigma>r"
    unfolding min_ichain.simps ichain.simps minimal_cond_def 
    unfolding INFM_disj_distrib[symmetric]
  proof (intro conjI)
    let ?S = "?QR O qrstep Q R O ?QR"
    let ?orig = "\<lambda> i. (s i, t i) \<in> P \<or> (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?S"
    let ?new = "\<lambda> i. (sr i, tr i) \<in> ?fP \<or> (tr i \<cdot> \<sigma>r i, sr (Suc i) \<cdot> \<sigma>r (Suc i)) \<in> ?S"
    from ichain 
    have inf: "(INFM i. ?orig i)" unfolding INFM_disj_distrib by simp
    show "(INFM i. ?new i)"
      unfolding INFM_nat_le
    proof
      fix m
      from inf[unfolded INFM_nat_le, rule_format, of "m + n"] obtain n' where n': "n' \<ge> m + n"
        and orig: "?orig n'" by blast
      from n' have "\<exists> k. n' = k + n \<and> k \<ge> m" by arith
      then obtain k where n': "n' = k + n" and k: "k \<ge> m" by blast
      from main[of k] main[of "Suc k"] orig[unfolded n'] have "?new k"
        by auto
      with k show "\<exists> k \<ge> m. ?new k" by blast
    qed
  qed (insert main ichain, auto)
  thus ?thesis by blast
qed


lemma instantiation_proc: fixes R Rw P Pw :: "('f,string)trs"
  assumes ecap: "is_ecap ecap"
  and fin: "finite_dpp (replace (s,t) sts P, replace (s,t) sts Pw, Q, R, Rw)"
  and main: "\<And> u v \<mu>. 
  (u,v) \<in> P \<union> Pw \<Longrightarrow> 
  mgu_class (ecap (R \<union> Rw) Q (mv_xvar ` {u}) (mv_xvar v)) s = Some \<mu> \<Longrightarrow> 
  mv_yvar s \<cdot> \<mu> \<in> NF_terms Q \<Longrightarrow>
  mv_xvar u \<cdot> \<mu> \<in> NF_terms Q \<Longrightarrow>
  \<exists> (s',t') \<in> sts. instance_rule (mv_yvar s \<cdot> \<mu>, mv_yvar t \<cdot> \<mu>) (s',t')"
  shows "finite_dpp (P,Pw,Q,R,Rw)"
  unfolding finite_dpp_def
proof (clarify)
  fix si ti \<sigma>i
  assume min_ichain: "min_ichain (P,Pw,Q,R,Rw) si ti \<sigma>i"
  note ichain = min_ichain[unfolded min_ichain.simps ichain.simps minimal_cond_def]
  let ?P = "\<lambda> P' i s' t' \<sigma>'. si i \<cdot> \<sigma>i i = s' \<cdot> \<sigma>' \<and> ti i \<cdot> \<sigma>i i = t' \<cdot> \<sigma>' \<and> (s',t') \<in> replace (s,t) sts P'"
  let ?QR = "(qrstep Q (R \<union> Rw))^*"
  {
    fix P' and i :: nat
    assume i: "i \<ge> Suc 0" and pair: "(si i, ti i) \<in> P'"
    have "\<exists> s' t' \<sigma>'. ?P P' i s' t' \<sigma>'"
    proof (cases "(si i, ti i) = (s,t)")
      case False
      with pair show ?thesis unfolding replace_def by auto
    next
      case True
      hence True: "si i = s" "ti i = t" by auto
      from i obtain j where i: "i = Suc j" by (cases i, auto)
      note inst = ecap_steps_mgu[OF ecap]
      from ichain have "(si j, ti j) \<in> P \<union> Pw" by auto
      note main = main[OF this]
      from ichain have NF1: "si j \<cdot> \<sigma>i j \<in> NF_terms Q" by simp
      from NF1 have "{si j} \<cdot>set \<sigma>i j \<subseteq> NF_terms Q" by simp
      note inst = inst[OF this]
      from ichain have "(ti j \<cdot> \<sigma>i j, si i \<cdot> \<sigma>i i) \<in> ?QR" unfolding i by simp
      with True have "(ti j \<cdot> \<sigma>i j, s \<cdot> \<sigma>i i) \<in> ?QR" by auto
      note inst = inst[OF this]
      from inst obtain \<mu> \<delta> where 
        mgu: "mgu_class (ecap (R \<union> Rw) Q (mv_xvar ` {si j}) (mv_xvar (ti j))) s = Some \<mu>"
        and \<sigma>j: "\<And> s. s \<cdot> \<sigma>i j = mv_xvar s \<cdot> \<mu> \<cdot> \<delta>"
        and \<sigma>i: "\<And> s. s \<cdot> \<sigma>i i = mv_yvar s \<cdot> \<mu> \<cdot> \<delta>" by blast
      note main = main[OF mgu]
      from ichain have "si i \<cdot> \<sigma>i i \<in> NF_terms Q" by simp
      hence NF2: "s \<cdot> \<sigma>i i \<in> NF_terms Q" unfolding True .
      from NF_instance[OF NF2[unfolded \<sigma>i]] have "mv_yvar s \<cdot> \<mu> \<in> NF_terms Q" .      
      note main = main[OF this]
      from NF_instance[OF NF1[unfolded \<sigma>j]] have "mv_xvar (si j) \<cdot> \<mu> \<in> NF_terms Q" .
      note main = main[OF this]      
      from main obtain s' t' where st': "(s',t') \<in> sts" and inst: "instance_rule (mv_yvar s \<cdot> \<mu>, mv_yvar t \<cdot> \<mu>) (s',t')" by auto
      from pair True st' have mem: "(s',t') \<in> replace (s,t) sts P'" unfolding replace_def by auto
      from inst[unfolded instance_rule_def] obtain \<tau> where id: 
        "mv_yvar s \<cdot> \<mu> = s' \<cdot> \<tau>" "mv_yvar t \<cdot> \<mu> = t' \<cdot> \<tau>" by auto        
      show ?thesis unfolding True \<sigma>i id
      proof (intro exI conjI)
        show "s' \<cdot> \<tau> \<cdot> \<delta> = s' \<cdot> (\<tau> \<circ>s \<delta>)" by simp
      qed (insert mem, auto)
    qed
  } note main = this
  let ?RP = "replace (s,t) sts P"
  let ?RPw = "replace (s,t) sts Pw"
  from ichain_instantiation[OF min_ichain main, of "Suc 0" "\<lambda> x. x"]
  have "\<exists> s' t' \<sigma>'. min_ichain (?RP, ?RPw, Q, R, Rw) s' t' \<sigma>'" by blast
  with fin show False unfolding finite_dpp_def by blast
qed

lemma forward_instantiation_proc: fixes R Rw P Pw :: "('f,string)trs" and ecap Q U
  defines urc: "urc \<equiv> is_ur_closed_term_mv' (R \<union> Rw) U Q ecap full_af"
  assumes ecap: "is_ecap ecap"
  and U_None: "U_opt = None \<Longrightarrow> U = R \<union> Rw"
  and U_Some: "U_opt = Some U \<Longrightarrow> vars_term t \<subseteq> vars_term s \<and> NF_terms Q \<subseteq> NF_trs (R \<union> Rw) \<and> urc {s} t \<and> (\<forall> (l,r) \<in> U. urc (set (args l)) r)"
  and U_opt: "U_opt = None \<or> U_opt = Some U"
  and fin: "finite_dpp (replace (s,t) sts P, replace (s,t) sts Pw, Q, R, Rw)"
  and main: "\<And> u v \<mu>. 
  (u,v) \<in> P \<union> Pw \<Longrightarrow> 
  mgu_class (ecap (U^-1) {} {} (mv_xvar u)) t = Some \<mu> \<Longrightarrow> 
  mv_yvar s \<cdot> \<mu> \<in> NF_terms Q \<Longrightarrow>
  mv_xvar u \<cdot> \<mu> \<in> NF_terms Q \<Longrightarrow>
  \<exists> (s',t') \<in> sts. instance_rule (mv_yvar s \<cdot> \<mu>, mv_yvar t \<cdot> \<mu>) (s',t')"
  shows "finite_dpp (P,Pw,Q,R,Rw)"
  unfolding finite_dpp_def
proof (clarify)
  fix si ti \<sigma>i
  assume min_ichain: "min_ichain (P,Pw,Q,R,Rw) si ti \<sigma>i"
  note ichain = min_ichain[unfolded min_ichain.simps ichain.simps minimal_cond_def]
  let ?P = "\<lambda> P' i s' t' \<sigma>'. si i \<cdot> \<sigma>i i = s' \<cdot> \<sigma>' \<and> ti i \<cdot> \<sigma>i i = t' \<cdot> \<sigma>' \<and> (s',t') \<in> replace (s,t) sts P'"
  let ?QR = "(qrstep Q (R \<union> Rw))^*"
  let ?QU = "(qrstep Q U)^*"
  let ?U = "(rstep U)^*"
  let ?U' = "(rstep (U^-1))^*"
  {
    fix P' and i :: nat
    assume pair: "(si i, ti i) \<in> P'"
    have "\<exists> s' t' \<sigma>'. ?P P' i s' t' \<sigma>'"
    proof (cases "(si i, ti i) = (s,t)")
      case False
      with pair show ?thesis unfolding replace_def by auto
    next
      case True
      hence True: "si i = s" "ti i = t" by auto
      note inst = ecap_steps_mgu[OF ecap]
      note inst = inst[where Q = "{}" and S = "{}", unfolded qrstep_rstep_conv]
      have id: "?U^-1 = ?U'" unfolding rtrancl_converse[symmetric] reverseTrs ..
      from ichain have "(ti i \<cdot> \<sigma>i i, si (Suc i) \<cdot> \<sigma>i (Suc i)) \<in> ?QR" by simp
      hence steps: "(t \<cdot> \<sigma>i i, si (Suc i) \<cdot> \<sigma>i (Suc i)) \<in> ?QR" (is "?ts \<in> _") unfolding True .
      from ichain have "si i \<cdot> \<sigma>i i \<in> NF_terms Q"  by auto
      from this[unfolded True] have NF: "s \<cdot> \<sigma>i i \<in> NF_terms Q" .
      from U_opt
      have "?ts \<in> ?QU"
      proof
        assume "U_opt = None"
        from U_None[OF this] steps show ?thesis by auto
      next
        assume "U_opt = Some U"
        from U_Some[OF this] have inn: "NF_terms Q \<subseteq> NF_trs (R \<union> Rw)"
          and urct: "urc {s} t"
          and urcU: "\<And> l r. (l,r) \<in> U \<Longrightarrow> urc (set (args l)) r" 
          and var_cond: "vars_term t \<subseteq> vars_term s" by auto
        interpret R_Q_U_ecap "R \<union> Rw" U Q ecap
          by (unfold_locales, rule ecap, rule inn)
        from NF have NF: "{s} \<cdot>set \<sigma>i i \<subseteq> NFQ" by auto
        from ichain have "SN_on (qrstep Q (R \<union> Rw)) {ti i \<cdot> \<sigma>i i}" by auto
        from this[unfolded True] have SN: "wf_trs (R \<union> Rw) \<or> SN_on (qrstep Q (R \<union> Rw)) {t \<cdot> \<sigma>i i}" ..
        note urc = is_ur_closed_term[OF steps _ NF urcU[unfolded urc] subset_refl SN urct[unfolded urc]]
        {
          fix u
          assume "u \<in> \<sigma>i i ` vars_term t"
          then obtain x where x: "x \<in> vars_term t" and u: "u = \<sigma>i i x" by auto
          from var_cond x have x: "x \<in> vars_term s" by auto
          hence "s \<unrhd> Var x" by auto
          hence "s \<cdot> \<sigma>i i \<unrhd> Var x \<cdot> \<sigma>i i" by auto
          hence "s \<cdot> \<sigma>i i \<unrhd> \<sigma>i i x" by simp
          from NF_subterm[OF _ this] NF 
          have "u \<in> NFQ" unfolding u by auto
        }
        hence "\<sigma>i i ` vars_term t \<subseteq> NFQ" by auto
        from urc[OF this] have "?ts \<in> (qrstep Q ((R \<union> Rw) \<inter> U))^*" .
        thus ?thesis using rtrancl_mono[OF qrstep_mono[OF _ subset_refl]] by blast
      qed
      hence "?ts \<in> ?U"
        using rtrancl_mono[OF qrstep_mono[OF subset_refl, of Q "{}" U]]
        by auto
      hence "(si (Suc i) \<cdot> \<sigma>i (Suc i), t \<cdot> \<sigma>i i) \<in> ?U'" unfolding id[symmetric] by simp
      note inst = inst[OF _ this]
      from inst obtain \<mu> \<delta> where 
        mgu: "mgu_class (ecap (U^-1) {} {} (mv_xvar (si (Suc i)))) t = Some \<mu>"
        and \<sigma>si: "\<And> s. s \<cdot> \<sigma>i (Suc i) = mv_xvar s \<cdot> \<mu> \<cdot> \<delta>"
        and \<sigma>i: "\<And> s. s \<cdot> \<sigma>i i = mv_yvar s \<cdot> \<mu> \<cdot> \<delta>" by auto
      from ichain have "(si (Suc i), ti (Suc i)) \<in> P \<union> Pw" by auto
      note main = main[OF this mgu]
      from NF_instance[OF NF[unfolded \<sigma>i]] have "mv_yvar s \<cdot> \<mu> \<in> NF_terms Q" .
      note main = main[OF this]
      from ichain have NF2: "si (Suc i) \<cdot> \<sigma>i (Suc i) \<in> NF_terms Q" by simp
      from NF_instance[OF NF2[unfolded \<sigma>si]] have "mv_xvar (si (Suc i)) \<cdot> \<mu> \<in> NF_terms Q" . 
      note main = main[OF this]
      from main obtain s' t' where st': "(s',t') \<in> sts" and inst: "instance_rule (mv_yvar s \<cdot> \<mu>, mv_yvar t \<cdot> \<mu>) (s',t') " by auto
      from pair True st' have mem: "(s',t') \<in> replace (s,t) sts P'" unfolding replace_def by auto
      from inst[unfolded instance_rule_def] obtain \<tau> where id: 
        "mv_yvar s \<cdot> \<mu> = s' \<cdot> \<tau>" "mv_yvar t \<cdot> \<mu> = t' \<cdot> \<tau>" by auto        
      show ?thesis unfolding True \<sigma>i id
      proof (intro exI conjI)
        show "s' \<cdot> \<tau> \<cdot> \<delta> = s' \<cdot> (\<tau> \<circ>s \<delta>)" by simp
      qed (insert mem, auto)
    qed
  } note main = this
  let ?RP = "replace (s,t) sts P"
  let ?RPw = "replace (s,t) sts Pw"
  from ichain_instantiation[OF min_ichain main, of "0" "\<lambda> x. x"]
  have "\<exists> s' t' \<sigma>'. min_ichain (?RP, ?RPw, Q, R, Rw) s' t' \<sigma>'" by blast
  with fin show False unfolding finite_dpp_def by blast
qed

end