(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Loops
imports
  Trs
  "~~/src/HOL/Library/Monad_Syntax"
  Substitution
begin


type_synonym ('f, 'v) redex_prob     = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) match_prob     = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) gmatch_prob    = "('f, 'v) match_prob list"
type_synonym ('f, 'v) ident_prob     = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) nident_prob    = "('f, 'v )term \<times> ('f, 'v) term \<times> nat"
type_synonym ('f, 'v) match_solution = "nat \<times> ('f, 'v) subst"
type_synonym ('f, 'v) redex_solution = "nat \<times> ('f, 'v) subst \<times> ('f, 'v) ctxt"

fun eroot :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) + 'v" where
  "eroot (Fun f ts) = Inl (f, length ts)"
| "eroot (Var x) = Inr x"

definition gmatch_solved_form :: "('f,'v)gmatch_prob \<Rightarrow> bool" where
  "gmatch_solved_form gmp \<equiv> \<forall>tl\<in>set gmp. is_Var (snd tl)"


locale fixed_subst = 
  fixes \<mu> :: "('f,'v)subst"
begin


definition redex_solution :: "('f,'v)redex_prob \<Rightarrow> ('f,'v)redex_solution \<Rightarrow> bool"
  where "redex_solution rp sol \<equiv> case (rp,sol) of ((t,l),(i,\<sigma>,C)) \<Rightarrow> t \<cdot> (\<mu> ^^ i) = C\<langle>l \<cdot> \<sigma>\<rangle>"

definition match_solution :: "('f,'v)match_prob \<Rightarrow> ('f,'v)match_solution \<Rightarrow> bool"
  where "match_solution mp sol \<equiv> case (mp,sol) of ((t,l),(i,\<sigma>)) \<Rightarrow> t \<cdot> (\<mu> ^^ i) = l \<cdot> \<sigma>"

definition gmatch_solution :: "('f,'v)gmatch_prob \<Rightarrow> ('f,'v)match_solution \<Rightarrow> bool"
  where "gmatch_solution gmp sol \<equiv> \<forall> tl \<in> set gmp. match_solution tl sol"

definition ident_solution :: "('f,'v)ident_prob \<Rightarrow> nat \<Rightarrow> bool"
  where "ident_solution ip n \<equiv> case ip of (s,t) \<Rightarrow> s \<cdot> (\<mu> ^^ n) = t \<cdot> (\<mu> ^^ n)"

definition W :: "('f,'v)term \<Rightarrow> 'v set" where "W t \<equiv> \<Union> {vars_term (t \<cdot> \<mu> ^^ i) | i :: nat. True}"

definition match_prob_of_rp :: "('f,'v)redex_prob \<Rightarrow> ('f,'v)match_prob set"
  where "match_prob_of_rp rp \<equiv> case rp of (t,Var x) \<Rightarrow> {(t,Var x)} | (t, Fun f ls) \<Rightarrow> 
             {(u,Fun f ls) | s u. s \<in> {t} \<union> \<mu> ` W t \<and> s \<unrhd> u \<and> is_Fun u}"


lemma match_prob_of_rp_sound: assumes sol: "match_solution (u,l') (i,\<sigma>)"
  and mp: "(u,l') \<in> match_prob_of_rp (t,l)"
  shows "\<exists> i C. redex_solution (t,l) (i,\<sigma>,C)"
proof -
  from sol[unfolded match_solution_def]
  have sol: "u \<cdot> \<mu> ^^ i = l' \<cdot> \<sigma>" by simp
  show ?thesis 
  proof (cases l)
    case (Var x)
    from mp[unfolded match_prob_of_rp_def] have "(u,l') = (t,l)" using Var by auto
    with sol have "t \<cdot> \<mu> ^^ i = \<box>\<langle>l\<cdot>\<sigma>\<rangle>" by simp
    thus ?thesis unfolding redex_solution_def split by blast
  next
    case (Fun f ls)
    from mp[unfolded match_prob_of_rp_def] obtain s where 
      s: "s \<in> {t} \<union> \<mu> ` W t" and su: "s \<unrhd> u" 
      and l': "l' = l" unfolding Fun by auto
    note sol = sol[unfolded l']
    from su obtain C where su: "s = C\<langle>u\<rangle>" ..
    from s have "\<exists> D n. t \<cdot> \<mu> ^^ n = D\<langle>s\<rangle>"
    proof
      assume s: "s \<in> {t}"
      show ?thesis by (rule exI[of _ \<box>], rule exI[of _ 0], insert s, auto)
    next
      assume "s \<in> \<mu> ` W t"
      then obtain x i where x: "x \<in> vars_term (t \<cdot> \<mu> ^^ i)" and s: "s = \<mu> x" unfolding W_def by auto
      from x have "\<exists> C. t \<cdot> \<mu> ^^ i = C\<langle>Var x\<rangle>" 
      proof (induct i arbitrary: x)
        case 0
        hence "x \<in> vars_term t" by simp
        from supteq_Var[OF this] obtain C where t: "t = C\<langle>Var x\<rangle>" ..          
        thus ?case by auto
      next
        case (Suc i)
        have id0: "t \<cdot> \<mu> ^^ Suc i = t \<cdot> \<mu> ^^ i \<cdot> \<mu>" unfolding subst_power_Suc by auto
        from Suc have "x \<in> vars_term (t \<cdot> \<mu> ^^ i \<cdot> \<mu>)" unfolding id0 by simp
        from this[unfolded vars_term_subst[of "t \<cdot> \<mu> ^^ i"]]
        obtain y where x: "x \<in> vars_term (\<mu> y)" and y: "y \<in> vars_term (t \<cdot> \<mu> ^^ i)" by auto
        from Suc(1)[OF y] obtain C where id1: "t \<cdot> \<mu> ^^ i = C\<langle>Var y\<rangle>" by auto
        from supteq_Var[OF x] obtain D where id2: "\<mu> y = D\<langle>Var x\<rangle>" ..
        show ?case 
          by (rule exI[of _ "(C \<cdot>c \<mu>) \<circ>c D"], unfold id0 id1, simp add: id2)
      qed
      then obtain C where id: "t \<cdot> \<mu> ^^ i = C\<langle>Var x\<rangle>" ..      
      show ?thesis
      proof (intro exI)
        have "t \<cdot> \<mu> ^^ (Suc i) = t \<cdot> \<mu> ^^ i \<cdot> \<mu>" unfolding subst_power_Suc by simp
        also have "... = (C \<cdot>c \<mu>)\<langle>\<mu> x\<rangle>" unfolding id by simp
        finally show "t \<cdot> \<mu> ^^ Suc i = (C \<cdot>c \<mu>)\<langle>s\<rangle>" unfolding s .
      qed
    qed
    then obtain D n where ts: "t \<cdot> \<mu> ^^ n = D\<langle>s\<rangle>" by blast
    let ?E = "D \<circ>c C"
    have "t \<cdot> (\<mu> ^^ (n + i)) = (t \<cdot> \<mu> ^^ n) \<cdot> \<mu> ^^ i" unfolding subst_power_compose_distrib
      by simp
    also have "... = ?E\<langle>u\<rangle> \<cdot> \<mu> ^^ i" unfolding ts su by simp
    also have "... = (?E \<cdot>c \<mu>^^i)\<langle> u \<cdot> \<mu> ^^ i\<rangle>" by simp
    finally show ?thesis unfolding sol redex_solution_def split by blast
  qed
qed

lemma match_prob_of_rp_complete: assumes sol: "redex_solution (t,l) (i,\<sigma>,C)"
  shows "\<exists> \<tau> u j. (u,l) \<in> match_prob_of_rp (t,l) \<and> match_solution (u,l) (j,\<tau>)"
proof -
  from sol[unfolded redex_solution_def]
  have sol: "t \<cdot> \<mu> ^^ i = C\<langle>l \<cdot> \<sigma>\<rangle>" by simp
  show ?thesis 
  proof (cases l)
    case (Var x)
    hence mem: "(t,Var x) \<in> match_prob_of_rp (t,Var x)" unfolding match_prob_of_rp_def by auto
    show ?thesis unfolding Var
    proof (intro exI conjI, rule mem, unfold match_solution_def split)
      show "t \<cdot> \<mu> ^^ 0 = Var x \<cdot> (\<lambda> _. t)" by simp
    qed
  next
    case (Fun f ls)    
    let ?mp = "\<lambda> t. {u. \<exists> s. s \<in> {t} \<union> \<mu> ` W t \<and> s \<unrhd> u \<and> is_Fun u}"
    have mps: "match_prob_of_rp (t,l) = (\<lambda> u. (u,l)) ` ?mp t" unfolding match_prob_of_rp_def Fun by auto
    let ?mpsol = "\<lambda> t u j. u \<in> ?mp t \<and> u \<cdot> \<mu> ^^ j = l \<cdot> \<sigma>"
    let ?ex = "\<lambda> t. \<exists> u j. ?mpsol t u j"
    let ?sol = "\<lambda> t i C. t \<cdot> \<mu> ^^ i = C\<langle>l \<cdot> \<sigma>\<rangle>"
    let ?p = "\<lambda> (t,i,C). ?sol t i C \<longrightarrow> ?ex t"
    let ?m = "\<lambda> (t,i,C). i + size C"
    have "?p (t,i,C)" 
    proof (induct rule: wf_induct[of _ ?p, OF wf_measure, of ?m])
      case (1 tiC)
      obtain t i C where tiC: "tiC = (t,i,C)"  by (cases tiC, auto)
      note 1 = 1[THEN spec, THEN mp, unfolded tiC]
      show ?case unfolding tiC split
      proof (intro impI)
        assume sol: "?sol t i C"
        show "?ex t"
        proof (cases t)
          case (Var x)
          from sol obtain j where i: "i = Suc j" unfolding Fun Var by (cases i, cases C, auto)
          have sol: "?sol (\<mu> x) j C" unfolding sol[symmetric] i Var by auto          
          let ?tiC = "(\<mu> x,j,C)"
          have "(?tiC,t,i,C) \<in> measure ?m" unfolding Var i by auto
          from 1[OF this, unfolded split, THEN mp, OF sol]
          obtain u k where sol: "?mpsol (\<mu> x) u k" by auto
          from sol have mem: "u \<in> ?mp (\<mu> x)" and id: "u \<cdot> \<mu> ^^ k = l \<cdot> \<sigma>" by auto
          from mem obtain s where s: "s \<in> {\<mu> x} \<union> \<mu> ` W (\<mu> x)" and mores: "s \<unrhd> u" "is_Fun u" by auto
          from s have s: "s \<in> \<mu> ` W t"
          proof 
            assume "s \<in> {\<mu> x}"
            hence s: "s = \<mu> x" by simp
            show "s \<in> \<mu> ` W t" unfolding W_def
              by (rule, rule s, unfold Var, rule, rule, rule exI[of _ 0], auto)
          next
            assume "s \<in> \<mu> ` W (\<mu> x)"
            from this[unfolded W_def] obtain i where s: "s \<in> \<mu> ` vars_term (\<mu> x \<cdot> \<mu> ^^ i)" by auto
            have id: "vars_term (\<mu> x \<cdot> \<mu> ^^ i) = vars_term (t \<cdot> \<mu> ^^ Suc i)" unfolding Var by simp
            show "s \<in> \<mu> ` W t" unfolding W_def using s[unfolded id] by blast
          qed
          show ?thesis
            by (intro exI, rule conjI[OF _ id], rule, rule exI[of _ s], insert s mores, auto)
        next
          case (Fun f ts)
          show ?thesis
          proof (cases C)
            case Hole
            from sol have sol: "t \<cdot> \<mu> ^^ i = l \<cdot> \<sigma>" unfolding Hole by auto
            show ?thesis
              by (intro exI, rule conjI[OF _ sol], rule, rule exI[of _ t], unfold Fun, auto)
          next
            case (More g bef D aft)
            let ?n = "length bef"
            from arg_cong[OF sol, of "length o args"] have len: "?n < length ts" unfolding Fun More by auto
            have "ts ! ?n \<cdot> \<mu> ^^ i = args (t \<cdot> \<mu> ^^ i) ! ?n" unfolding Fun 
              by (simp add: nth_map[OF len])
            also have "... = args (C\<langle>l\<cdot>\<sigma>\<rangle>) ! ?n" using sol by simp
            also have "... = D\<langle>l\<cdot>\<sigma>\<rangle>" unfolding More by simp
            finally have sol: "?sol (ts ! ?n) i D" .
            have "((ts ! ?n, i, D), t, i, C) \<in> measure ?m" unfolding More by simp
            from 1[OF this, unfolded split, THEN mp, OF sol]
            obtain u k where sol: "?mpsol (ts ! ?n) u k" by auto
            from sol have mem: "u \<in> ?mp (ts ! ?n)" and id: "u \<cdot> \<mu> ^^ k = l \<cdot> \<sigma>" by auto
            from mem obtain s where s: "s \<in> {ts ! ?n} \<union> \<mu> ` W (ts ! ?n)" and mores: "s \<unrhd> u" "is_Fun u" by auto
            from len have mem: "ts ! ?n \<in> set ts" by auto
            from s have u: "u \<in> ?mp t" 
            proof
              assume "s \<in> {ts ! ?n}"
              hence s: "s = ts ! ?n" by simp
              show ?thesis
              proof (rule, rule exI[of _ t], intro conjI)
                show "t \<unrhd> u" unfolding Fun using mem mores(1)
                  unfolding s by auto
              qed (insert mores, auto)
            next
              assume "s \<in> \<mu> ` W (ts ! ?n)"
              from this[unfolded W_def] obtain i where s: "s \<in> \<mu> ` vars_term (ts ! ?n \<cdot> \<mu> ^^ i)" by auto
              from mem have "vars_term (ts ! ?n) \<subseteq> vars_term t" unfolding Fun by auto
              with s have "s \<in> \<mu> ` vars_term (t \<cdot> \<mu> ^^ i)" unfolding vars_term_subst by auto
              hence "s \<in> \<mu> ` W t" unfolding W_def by auto
              with mores show ?thesis by auto
            qed
            with id have "?mpsol t u k" by auto
            thus ?thesis by blast
          qed
        qed
      qed
    qed                
    with sol show ?thesis unfolding mps match_solution_def by blast
  qed
qed

definition V_incr :: "'v set" where "V_incr \<equiv> {x . (\<exists> i . is_Fun (Var x \<cdot> \<mu> ^^ i))}"


lemma subst_V_incr: assumes \<mu>: "\<mu> x = Var y" shows "(x \<in> V_incr) = (y \<in> V_incr)"
proof (cases "x \<in> V_incr")
  case True
  then obtain n where n: "is_Fun (Var x \<cdot> \<mu> ^^ n)" unfolding V_incr_def by auto
  then obtain n where n: "is_Fun (Var x \<cdot> \<mu> ^^ (Suc n))" by (cases n, auto)
  hence n': "is_Fun (Var y \<cdot> \<mu> ^^ n)" using \<mu> by auto
  from n n' show ?thesis unfolding V_incr_def by blast
next
  case False
  have "y \<notin> V_incr"
  proof
    assume "y \<in> V_incr"
    then obtain n where n: "is_Fun (Var y \<cdot> \<mu> ^^ n)" unfolding V_incr_def by auto
    hence n': "is_Fun (Var x \<cdot> \<mu> ^^ Suc n)" using \<mu> by auto
    with False show False unfolding V_incr_def by blast
  qed
  with False show ?thesis by simp
qed

function
  simplify_mp ::
    "('f, 'v) gmatch_prob \<Rightarrow> ('f, 'v) gmatch_prob \<Rightarrow> (('f, 'v) gmatch_prob \<times> nat) option"
where
  "simplify_mp [] solved = Some (solved, 0)"
| "simplify_mp ((s, Var x) # mp) solved = simplify_mp mp ((s,Var x) # solved)"
| "simplify_mp ((Fun g ts, Fun f ls) # mp) solved = do {
    guard (f = g);
    pairs \<leftarrow> zip_option ts ls;
    simplify_mp (pairs @ mp) solved
  }"
| "simplify_mp ((Var x, Fun f ls) # mp) solved = do {
    guard (x \<in> V_incr);
    let m = map (\<lambda>(s, l). (s \<cdot> \<mu>, l));
    (smp,i) \<leftarrow> simplify_mp (m ((Var x, Fun f ls) # mp)) (m solved); 
    Some (smp, Suc i) 
  }"
  by pat_completeness auto
termination
proof -
  let ?nv = "\<lambda> x i. is_Fun (Var x \<cdot> \<mu> ^^ i)"
  let ?L = "\<lambda> x. LEAST i. ?nv x i"
  let ?f = "\<lambda> (t,l). case t of Fun _ _ \<Rightarrow> 0 | Var x \<Rightarrow> (if x \<in> V_incr then ?L x else 0)"
  let ?R = "measures [\<lambda> (mp,sp). list_size size (map snd (mp @ sp)), 
                      \<lambda>(mp,sp). list_size ?f (mp @ sp),
                      \<lambda>(mp,sp). length mp]"
  show ?thesis
  proof 
    show "wf ?R" by (rule wf_measures)
  next
    fix s x mp solved
    show "((mp,(s,Var x) # solved), (s, Var x) # mp, solved) \<in> ?R"
      by (rule measures_lesseq[OF _ measures_lesseq[OF _ measures_less]], simp_all)
  next
    fix g::"'f"
      and ts::"('f, 'v) term list" and f ls mp solved and pairs :: "('f,'v) gmatch_prob"
    assume "f = g" and d: "zip_option ts ls = Some pairs"
    have tedious: "list_size (size o snd) (zip ts ls) < Suc (Suc (list_size size ls))"
    proof (induct ls arbitrary: ts)
      case Nil thus ?case by auto
    next
      case (Cons l ls ts)
      thus ?case by (cases ts, auto)
    qed
    show "((pairs @ mp, solved), (Fun g ts, Fun f ls) # mp, solved) \<in> ?R"
      by (rule measures_less) (insert d tedious, auto)
  next
    fix x f and ls :: "('f,'v)term list" and mp solved :: "('f,'v)gmatch_prob"
      and mu :: "('f,'v)gmatch_prob \<Rightarrow> ('f,'v)gmatch_prob"
    assume x: "x \<in> V_incr" and mu: "mu = map (\<lambda>(s,y). (s \<cdot> \<mu>, y))"
    let ?mu = "(\<lambda>(s,y). (s \<cdot> \<mu>, y))"
    have id: "size o snd o ?mu = size o snd"
      by (intro ext, auto)
    from x[unfolded V_incr_def] obtain n where "?nv x n" by auto
    from LeastI[of "?nv x", OF this] have nv: "?nv x (?L x)" .
    then obtain n where L: "?L x = Suc n" by (cases "?L x", auto)
    have "?f (?mu (Var x, Fun f ls)) < ?L x"
    proof (cases "\<mu> x")
      case (Fun f ts)
      with L show ?thesis by auto
    next
      case (Var y)
      from nv have ny: "?nv y n" unfolding L using Var by auto
      from Least_le[of "?nv y", OF ny] have Ly: "?L y \<le> n" .
      from ny have "y \<in> V_incr" unfolding V_incr_def by auto
      with Ly show ?thesis unfolding L using Var by auto
    qed            
    also have "... = ?f ((Var x, Fun f ls))" using x by simp
    finally have one: "?f (?mu (Var x, Fun f ls)) < ?f ((Var x, Fun f ls))" .
    have two: "list_size ?f (map ?mu (mp @ solved)) \<le> list_size ?f (mp @ solved)"
      unfolding list_size_map
    proof (rule list_size_pointwise)
      fix pair
      show "(?f o ?mu) pair \<le> ?f pair"
      proof (cases pair)
        case (Pair t l)
        show ?thesis 
        proof (cases t)
          case (Fun f ts)
          show ?thesis unfolding Pair Fun by simp
        next
          case (Var x) note x = this
          show ?thesis 
          proof (cases "\<mu> x")
            case (Fun f ts)
            show ?thesis unfolding Pair Var using Fun by auto
          next
            case (Var y)
            show ?thesis 
            proof (cases "y \<in> V_incr")
              case False
              with subst_V_incr[OF Var] have "x \<notin> V_incr" by auto
              with False show ?thesis unfolding Pair using x Var by auto
            next
              case True
              with subst_V_incr[OF Var] have xi: "x \<in> V_incr" by auto
              then obtain n where "?nv x n" unfolding V_incr_def by auto
              from LeastI[of "?nv x", OF this] have nx: "?nv x (?L x)" .
              then obtain n where L: "?L x = Suc n" by (cases "?L x", auto)
              from nx have "?nv y n" unfolding L using Var by auto
              from Least_le[of "?nv y", OF this] have "?L y \<le> n" by auto
              with True show ?thesis unfolding Pair x using xi Var L by simp
            qed
          qed
        qed
      qed
    qed
    from one two 
    have main: "list_size ?f (map ?mu ((Var x, Fun f ls) # mp @ solved)) < list_size ?f ((Var x, Fun f ls) # mp @ solved)" 
      by auto
    show "((mu ((Var x, Fun f ls) # mp), mu solved), (Var x, Fun f ls) # mp, solved) \<in> ?R" unfolding mu
      by (rule measures_lesseq[OF _ measures_less], simp add: id, insert main, auto)
  qed
qed

lemma simplify_mp_solved_form: assumes 
  res: "simplify_mp mp1 mp2 = Some (mp3,i)" 
  and sf: "gmatch_solved_form mp2"
  shows "gmatch_solved_form mp3"
proof -
  let ?P = "\<lambda> mp1 mp2. \<forall> mp3i. simplify_mp mp1 mp2 = Some mp3i \<longrightarrow> gmatch_solved_form mp2 \<longrightarrow> gmatch_solved_form (fst mp3i)"
  have "?P mp1 mp2"
  proof (induct rule: simplify_mp.induct[of ?P])
    case (1 mp2)
    thus ?case by simp
  next
    case (2 s x mp1 mp2)
    show ?case
    proof (intro allI impI)
      fix mp3
      assume res: "simplify_mp ((s, Var x) # mp1) mp2 = Some mp3" and 
        sf: "gmatch_solved_form mp2"
      show "gmatch_solved_form (fst mp3)"
      proof (rule 2[THEN spec[of _ mp3], THEN mp, THEN mp])
        show "simplify_mp mp1 ((s, Var x) # mp2) = Some mp3"
          using res by simp
      next
        show "gmatch_solved_form ((s, Var x) # mp2)" using sf unfolding gmatch_solved_form_def
          by auto
      qed
    qed
  next
    case (3 g ts f ls mp1 mp2)
    show ?case
    proof (intro allI impI)
      fix mp3
      assume res: "simplify_mp ((Fun g ts, Fun f ls) # mp1) mp2 = Some mp3"
        and sf: "gmatch_solved_form mp2"
      hence fg: "f = g" by (cases "f = g", auto)
      note res = res[unfolded fg]
      from res obtain pairs where dec: "zip_option ts ls = Some pairs"
        by (cases "zip_option ts ls", auto)
      from res dec have res: "simplify_mp (pairs @ mp1) mp2 = Some mp3" by auto
      show "gmatch_solved_form (fst mp3)"
        using 3[OF fg dec] sf res by auto
    qed
  next
    case (4 x f ls mp1 mp2)
    show ?case
    proof (intro allI impI)
      fix mp3
      assume res: "simplify_mp ((Var x, Fun f ls) # mp1) mp2 = Some mp3"
        and sf: "gmatch_solved_form mp2"
      hence x: "x \<in> V_incr" by (cases "x \<in> V_incr", auto)
      let ?mu = "map (\<lambda>(s,y). (s \<cdot> \<mu>, y))"
      from sf have sf: "gmatch_solved_form (?mu mp2)" unfolding gmatch_solved_form_def by auto
      let ?l = "simplify_mp (?mu ((Var x, Fun f ls) # mp1)) (?mu mp2)"
      obtain res where l: "?l = Some res" using res x by (cases ?l, auto)
      from res x l have "fst res = fst mp3" by (cases res, auto)
      with 4[OF x, THEN spec, THEN mp[OF _ l]] sf 
        show "gmatch_solved_form (fst mp3)" by auto
    qed
  qed
  with assms
  show ?thesis by simp
qed

lemma simplify_mp_solution: assumes "simplify_mp mp1 mp2 = mp3i"
  shows "(mp3i = None \<longrightarrow> \<not> gmatch_solution (mp1 @ mp2) (n,\<sigma>)) \<and> (\<forall> mp3 i. mp3i = Some (mp3,i) \<longrightarrow> gmatch_solution (mp1 @ mp2) (n + i, \<sigma>) = gmatch_solution mp3 (n,\<sigma>) \<and> (\<forall> k. k < i \<longrightarrow> \<not> gmatch_solution (mp1 @ mp2) (k,\<sigma>)))"
proof -
  let ?sol = "\<lambda> mp n. gmatch_solution mp (n, \<sigma>)"
  let ?eq = "\<lambda> mp1 mp2 mp3 n i. ?sol (mp1 @ mp2) (n + i) = ?sol mp3 n"
  let ?small = "\<lambda> mp1 mp2 i. \<forall> k. (k < i \<longrightarrow> \<not> ?sol (mp1 @ mp2) k)"
  let ?snd = "\<lambda> mp1 mp2 mp3i n. \<forall> mp3 i. mp3i = Some (mp3,i) \<longrightarrow> ?eq mp1 mp2 mp3 n i \<and> ?small mp1 mp2 i"
  let ?fst = "\<lambda> mp1 mp2 mp3i n. mp3i = None \<longrightarrow> \<not> ?sol (mp1 @ mp2) n"
  let ?both = "\<lambda> mp1 mp2 mp3i n. ?fst mp1 mp2 mp3i n \<and> ?snd mp1 mp2 mp3i n"
  let ?res = "\<lambda> mp1 mp2 mp3i. simplify_mp mp1 mp2 = mp3i"
  let ?P = "\<lambda> mp1 mp2. \<forall> mp3i n. ?res mp1 mp2 mp3i \<longrightarrow> ?both mp1 mp2 mp3i n"
  have "?P mp1 mp2" 
  proof (induct rule: simplify_mp.induct[of ?P])
    case (1 mp2)
    show ?case by simp
  next
    case (2 s x mp1 mp2)
    show ?case
    proof (intro allI impI)
      fix mp3i n
      assume "?res ((s,Var x) # mp1) mp2 mp3i"
      hence "?res mp1 ((s,Var x) # mp2) mp3i" by auto
      from 2[THEN spec, THEN spec, THEN mp[OF _ this]]
      show "?both ((s,Var x) # mp1) mp2 mp3i n"
        unfolding gmatch_solution_def by auto
    qed
  next
    case (3 g ts f ls mp1 mp2)
    show ?case
    proof (intro allI impI)
      fix mp3i n
      assume res: "?res ((Fun g ts, Fun f ls) # mp1) mp2 mp3i"
      show "?both ((Fun g ts, Fun f ls) # mp1) mp2 mp3i n"
      proof (cases "f = g")
        case False
        thus ?thesis using res unfolding gmatch_solution_def match_solution_def by auto
      next
        case True
        show ?thesis
        proof (cases "zip_option ts ls")
          case None
          {
            fix f g
            have "map f ts \<noteq> map g ls" 
            proof 
              assume "map f ts = map g ls"
              from arg_cong[OF this, of length] and None show False by simp
            qed
          } 
          with True None res show ?thesis
            unfolding gmatch_solution_def match_solution_def 
            by (auto)
        next
          case (Some pairs)
          from Some have len: "length ts = length ls" by auto
          from Some have pairs: "pairs = zip ts ls" by auto
          from res have res: "?res (pairs @ mp1) mp2 mp3i" using True Some by auto
          from len have len: "length ts = length (map (\<lambda> t. t \<cdot> \<sigma>) ls)" by simp
          have main: "\<And> n. match_solution (Fun g ts, Fun g ls) (n,\<sigma>) = gmatch_solution (zip ts ls) (n,\<sigma>)" 
            unfolding gmatch_solution_def match_solution_def split set_zip subst_apply_term.simps term.simps
            unfolding map_nth_eq_conv[OF len]
            by (insert len, auto)
          note rec = 3[OF True Some, THEN spec, THEN spec, THEN mp[OF _ res], of n]
          show ?thesis
          proof (cases mp3i)
            case None
            with rec show ?thesis using main unfolding True pairs gmatch_solution_def by auto
          next
            case (Some mp3)
            then obtain mp i where mp3i: "mp3i = Some (mp,i)" by (cases mp3, auto)
            with rec show ?thesis unfolding True pairs
              by (auto simp: main gmatch_solution_def)
          qed
        qed
      qed
    qed
  next
    case (4 x f ls mp1 mp2)
    show ?case
    proof (intro allI impI)
      fix mp3i n
      assume res: "?res ((Var x, Fun f ls) # mp1) mp2 mp3i"
      show "?both ((Var x, Fun f ls) # mp1) mp2 mp3i n"
      proof (cases "x \<in> V_incr")
        case False
        have "\<not> match_solution (Var x, Fun f ls) (n, \<sigma>)"
        proof
          assume "match_solution (Var x, Fun f ls) (n, \<sigma>)"
          hence "Var x \<cdot> (\<mu> ^^ n) = Fun f ls \<cdot> \<sigma>" 
            unfolding match_solution_def by auto
          hence "is_Fun (Var x \<cdot> (\<mu> ^^ n))" by auto
          with False show False unfolding V_incr_def by auto
        qed
        with res False show ?thesis unfolding gmatch_solution_def by auto
      next
        case True
        let ?mp1 = "((Var x, Fun f ls) # mp1)"
        let ?mu = "(\<lambda>(s,y). (s \<cdot> \<mu>, y))"
        let ?mmp1 = "map ?mu ?mp1"
        let ?mmp2 = "map ?mu mp2"
        let ?f = "(\<lambda>(smp, i). Some (smp, Suc i))"
        from True res 
        have res: "simplify_mp ?mmp1 ?mmp2 \<guillemotright>= ?f = mp3i" by (simp add: Let_def)
        obtain mp3i' where mp3i': "simplify_mp ?mmp1 ?mmp2 = mp3i'" by auto
        with res have res: "mp3i' \<guillemotright>= ?f = mp3i" by simp
        obtain pairs where pairs: "pairs = ?mp1 @ mp2" by auto
        have id: "?both ((Var x, Fun f ls) # mp1) mp2 mp3i n = (?both pairs [] mp3i n)" unfolding pairs gmatch_solution_def by auto
        from 4[OF True refl, THEN spec, THEN spec, THEN mp[OF _ mp3i']]
          have rec: "\<And> n. ?both (map ?mu pairs) [] mp3i' n"
          unfolding pairs gmatch_solution_def by auto
        show ?thesis unfolding id
        proof (cases mp3i')
          case None
          with res have res: "mp3i = None" by auto
          have "\<not> ?sol pairs n" 
          proof 
            assume sol: "?sol pairs n"
            then obtain m where n: "n = Suc m" unfolding gmatch_solution_def match_solution_def pairs
              by (cases n, auto)
            from sol[unfolded n] have "?sol (map ?mu pairs) m"
              unfolding gmatch_solution_def match_solution_def pairs by auto
            with rec[of m] None show False by auto
          qed
          with res show "?both pairs [] mp3i n" by auto
        next
          case (Some mpi')
          then obtain mp i where mpi': "mpi' = (mp,i)" by (cases mpi', auto)
          with Some have mp3i': "mp3i' = Some (mp,i)" by auto
          with res have mp3i: "mp3i = Some (mp, Suc i)" by auto
          from rec have id: "\<And> n. ?sol (map ?mu pairs) (n + i) = ?sol mp n" unfolding mp3i' by auto
          from rec have not: "\<And> k. k < i \<Longrightarrow> \<not> ?sol (map ?mu pairs) k" unfolding mp3i' by auto
          {
            fix k
            assume k: "k < Suc i"
            have "\<not> ?sol pairs k"
            proof (cases k)
              case (Suc l)
              from not[of l] k show ?thesis unfolding Suc gmatch_solution_def match_solution_def by auto
            next
              case 0
              thus ?thesis unfolding gmatch_solution_def match_solution_def pairs by auto
            qed
          } note not = this
          have id: "\<And> n. ?sol mp n = ?sol pairs (Suc n + i)" unfolding id[symmetric]
            unfolding gmatch_solution_def match_solution_def by auto
          thus "?both pairs [] mp3i n" unfolding mp3i using not by auto                
        qed
      qed
    qed
  qed
  with assms show ?thesis by auto
qed

lemma simplify_mp_complete:
  assumes "gmatch_solution mp (n,\<sigma>)"
  shows "\<exists> smp i k. simplify_mp mp [] = Some (smp,i) \<and> n = i + k \<and> gmatch_solution smp (k,\<sigma>)"
proof(cases "simplify_mp mp []")
  case None
  from simplify_mp_solution[OF None] assms
  show ?thesis by simp
next
  case (Some smpi)
  then obtain smp i where s: "simplify_mp mp [] = Some (smp,i)" by (cases smpi, auto)
  note main = simplify_mp_solution[OF s, THEN conjunct2, THEN spec, THEN spec, THEN mp[OF _ refl], of _ \<sigma>]
  show ?thesis
  proof (cases "n < i")
    case True
    with main assms show ?thesis by auto
  next
    case False
    hence "n = i + (n - i)" by auto
    then obtain k where n: "n = i + k" by auto
    hence n': "n = k + i" by simp
    show ?thesis
    proof (intro exI conjI, rule s, rule n)
      show "gmatch_solution smp (k,\<sigma>)" using main[of k] assms[unfolded n'] by simp
    qed
  qed
qed

lemma simplify_mp_sound: 
  assumes res: "simplify_mp mp [] = Some (smp,i)" 
  and sol: "gmatch_solution smp (n,\<sigma>)"
  shows "gmatch_solution mp (n + i, \<sigma>)"
  using simplify_mp_solution[OF res] sol by auto
end

fun ident_prob_of_smp :: "('f,'v)gmatch_prob \<Rightarrow> ('f,'v)ident_prob list"
  where "ident_prob_of_smp [] = []"
      | "ident_prob_of_smp ((t,Var x) # other) = map (\<lambda> (s,_). (t,s)) (filter (\<lambda> (_, s). s = Var x) other) @ ident_prob_of_smp other"

lemma ident_prob_of_smp: assumes "gmatch_solved_form smp"
  shows "set (ident_prob_of_smp smp) = { (fst (smp ! i), fst (smp ! j)) | i j. i < j \<and> j < length smp \<and> snd (smp ! i) = snd (smp ! j)}" (is "_ = ?s smp")
  using assms 
proof (induct smp)
  case (Cons tx smp)
  obtain t x where tx: "tx = (t,x)" by (cases tx, auto)
  from Cons(2) have var: "is_Var x" and sf: "gmatch_solved_form smp"
    unfolding tx gmatch_solved_form_def by auto
  from tx var obtain x where tx: "tx = (t, Var x)" by (cases x, auto)
  let ?one = "(\<lambda>(s,u). (t,s)) ` {y \<in> set smp. case y of (u,s) \<Rightarrow> s = Var x}"
  let ?p = "\<lambda> i j xs uv. i < j \<and> j < length xs \<and> snd (xs ! i) = snd (xs ! j) \<and> uv = (fst (xs ! i), fst (xs ! j))"
  have "set (ident_prob_of_smp (tx # smp)) = ?one \<union> ?s smp" 
    unfolding tx using Cons(1)[OF sf] by simp
  also have "... = ?s ((t,Var x)  # smp)" (is "?l = ?r")    
  proof -
    {
      fix u v 
      assume "(u,v) \<in> ?one"
      hence mem: "(v,Var x) \<in> set smp" and u: "u = t" by auto
      then obtain i where mem: "smp ! i = (v,Var x)" and i: "i < length smp" unfolding set_conv_nth by auto
      have "(u,v) \<in> ?r"
        by (rule, rule exI[of _ 0], rule exI[of _ "Suc i"], insert mem i u, auto)
    } hence "?one \<subseteq> ?r" ..
    moreover
    { 
      fix u v
      assume "(u,v) \<in> ?s smp"
      then obtain i j where p: "?p i j smp (u,v)" by auto
      have "(u,v) \<in> ?r"
        by (rule, rule exI[of _ "Suc i"], rule exI[of _ "Suc j"], insert p, auto)
    } hence "?s smp \<subseteq> ?r" ..
    moreover
    {
      fix u v 
      assume "(u,v) \<in> ?r"
      then obtain i j where p: "?p i j ((t,Var x) # smp) (u,v)" by auto
      from p obtain j' where j: "j = Suc j'" by (cases j, auto)
      have "(u,v) \<in> ?l"
      proof (cases i)
        case (Suc i')
        have "(u,v) \<in> ?s smp"
          by (rule, rule exI[of _ i'], rule exI[of _ j'], insert p Suc j, auto)
        thus ?thesis by simp
      next
        case 0
        have "(u,v) \<in> ?one"
          unfolding set_conv_nth using p unfolding j 0 by force
        thus ?thesis by simp
      qed
    } hence "?r \<subseteq> ?l" ..
    ultimately show ?thesis by blast
  qed
  finally show ?case unfolding tx .
qed simp

context fixed_subst
begin
lemma ident_prob_of_smp_complete: assumes sf: "gmatch_solved_form smp"
  and sol: "gmatch_solution smp (n,\<sigma>)"
  and mem: "(s,t) \<in> set (ident_prob_of_smp smp)"
  shows "ident_solution (s,t) n"
proof -
  from mem[unfolded ident_prob_of_smp[OF sf]]
  obtain i j where ij: "i < j" "j < length smp" 
    and snd: "snd (smp ! i) = snd (smp ! j)" 
    and fst: "s = fst (smp ! i)" "t = fst (smp ! j)" by auto
  from ij have ij: "i < length smp" "j < length smp" by auto
  hence "smp ! i \<in> set smp" "smp ! j \<in> set smp" by auto
  with sol[unfolded gmatch_solution_def] 
  have sol: "match_solution (smp ! i) (n,\<sigma>)" "match_solution (smp ! j) (n,\<sigma>)" by auto
  from fst obtain si where si: "smp ! i = (s,si)" by (cases "smp ! i", auto)
  from fst obtain sj where sj: "smp ! j = (t,sj)" by (cases "smp ! j", auto)
  from snd si sj have sj: "smp ! j = (t,si)" by auto
  from sol[unfolded si sj] show ?thesis unfolding match_solution_def ident_solution_def by auto
qed
  

lemma ident_prob_of_smp_sound: 
  assumes smp: "gmatch_solved_form smp"
  and sol: "\<And> s t. (s,t) \<in> set (ident_prob_of_smp smp) \<Longrightarrow> \<exists> m \<le> n. ident_solution (s,t) m"
  shows "\<exists> \<sigma>. gmatch_solution smp (n, \<sigma>)"
  unfolding gmatch_solution_def
proof (rule exI, clarify)
  fix t x
  assume mem: "(t,x) \<in> set smp"
  with smp[unfolded gmatch_solved_form_def] obtain x' where x: "x = Var x'" by (cases x, auto)
  let ?smp = "(map (\<lambda>(t,x). (the_Var x, t)) smp)"
  let ?\<sigma> = "\<lambda> x. the (lookup x ?smp) \<cdot> \<mu> ^^ n" 
  from mem x have mem': "(x',t) \<in> set ?smp" by force
  have "t \<cdot> \<mu> ^^ n = ?\<sigma> x'" 
  proof (cases "lookup x' ?smp")
    case None
    from lookup_complete[OF this] mem' have False by auto
    thus ?thesis by auto
  next
    case (Some s)
    from lookup_sound[OF this] obtain y' where me: "(s,y') \<in> set smp"
      and "the_Var y' = x'" by auto
    with smp[unfolded gmatch_solved_form_def] have y': "y' = Var x'" by (cases y', auto)
    with me have me: "(s,Var x') \<in> set smp" by auto
    from mem[unfolded x] me obtain i j where ij: "i < length smp" "j < length smp"
      and smpij: "smp ! i = (t,Var x')" "smp ! j = (s,Var x')" unfolding set_conv_nth by auto
    note sol = sol[unfolded ident_prob_of_smp[OF smp]]
    show ?thesis unfolding Some the.simps 
    proof (cases "i = j")
      case True
      with smpij have "s = t" by simp
      thus "t \<cdot> \<mu> ^^ n = s \<cdot> \<mu> ^^ n" by simp
    next
      case False
      hence "i < j \<or> j < i" by arith
      hence "\<exists> m \<le> n. t \<cdot> \<mu> ^^ m = s \<cdot> \<mu> ^^ m"
      proof
        assume lt: "i < j"
        have "\<exists> m \<le> n. ident_solution (t,s) m"
          by (rule sol, insert lt ij smpij, force)
        thus ?thesis unfolding ident_solution_def by auto
      next
        assume lt: "j < i"
        have "\<exists> m \<le> n. ident_solution (s,t) m"
          by (rule sol, insert lt ij smpij, force)
        thus ?thesis unfolding ident_solution_def by auto
      qed
      then obtain m where m: "m \<le> n" and id: "t \<cdot> \<mu> ^^ m = s \<cdot> \<mu> ^^ m" by auto
      from m have "n = m + (n - m)" by auto
      then obtain k where n: "n = m + k" by auto
      show "t \<cdot> \<mu> ^^ n = s \<cdot> \<mu> ^^ n" unfolding n subst_power_compose_distrib
        using id by simp
    qed
  qed
  thus "match_solution (t,x) (n, ?\<sigma>)" 
    unfolding match_solution_def split x by simp
qed


definition match_to_idents :: "('f,'v)match_prob \<Rightarrow> (nat \<times> ('f,'v)ident_prob list) option"
  where "match_to_idents t_l \<equiv> do {
       (smp,i) \<leftarrow> simplify_mp [t_l] [];
       Some (i,ident_prob_of_smp smp)
     }"

lemma match_to_idents_complete: assumes sol: "match_solution (t,l) (n,\<sigma>)"
  shows "\<exists> i idps. match_to_idents (t,l) = Some (i,idps) \<and> (\<forall> idp \<in> set idps. \<exists> n. ident_solution idp n)"
proof -
  from sol have "gmatch_solution [(t,l)] (n,\<sigma>)" unfolding gmatch_solution_def by auto
  from simplify_mp_complete[OF this] obtain smp i k
    where res: "simplify_mp [(t,l)] [] = Some (smp, i)" and n: "n = i + k" 
    and sol: "gmatch_solution smp (k,\<sigma>)" by auto
  from simplify_mp_solved_form[OF res] have sf: "gmatch_solved_form smp"
    unfolding gmatch_solved_form_def by auto
  from ident_prob_of_smp_complete[OF sf sol]
  show ?thesis
    by (force simp: match_to_idents_def res)
qed

lemma match_to_idents_sound: 
  assumes res: "match_to_idents (t,l) = Some (i,idps)"
  and sol: "\<And> idp. idp \<in> set idps \<Longrightarrow> ident_solution idp (ns idp)"
  shows "\<exists> \<sigma>. match_solution (t,l) (max_list (map ns idps) + i, \<sigma>)"
proof -
  from sol have  sol: "\<And> idp. idp \<in> set idps \<Longrightarrow> ident_solution idp (ns idp)" by auto
  let ?n = "max_list (map ns idps)"
  note res = res[unfolded match_to_idents_def]
  from res obtain smpi where res1: "simplify_mp [(t,l)] [] = Some smpi"
    by (cases "simplify_mp [(t,l)] []", auto)
  obtain smp j where smpi: "smpi = (smp,j)" by (cases smpi, auto)
  note res1 = res1[unfolded smpi]
  from simplify_mp_solved_form[OF res1] have sf: "gmatch_solved_form smp"
    unfolding gmatch_solved_form_def by auto
  note res = res[unfolded res1]
  from res have res2: "ident_prob_of_smp smp = idps"  and j: "j = i" by auto
  have sol: "\<exists> \<sigma>. gmatch_solution smp (?n, \<sigma>)"
  proof (rule ident_prob_of_smp_sound[OF sf, unfolded res2])
    fix s t
    assume mem: "(s,t) \<in> set idps"
    show "\<exists> m \<le> ?n. ident_solution (s,t) m"
      by (rule, rule conjI[OF max_list sol[OF mem]], insert mem, auto)
  qed
  then obtain \<sigma> where sol: "gmatch_solution smp (?n,\<sigma>)" by auto
  from simplify_mp_sound[OF res1[unfolded smpi] sol]
  show ?thesis unfolding j gmatch_solution_def by auto
qed
end

section {* solving identity problems *}

context fixed_subst
begin

declare size_simps[simp]
fun conflict_pos :: "('f,'v)term \<times> ('f,'v)term \<times> nat \<Rightarrow> (('f,'v)term \<times> ('f,'v)term \<times> nat) list"
  where "conflict_pos (s,Var x,Suc n) = conflict_pos (s,\<mu> x, n)"
    |   "conflict_pos (Var x,Var y,0) = (if x = y then [] else [(Var x, Var y,0)])"
    |   "conflict_pos (Fun f ts,Var y,0) = [(Var y, Fun f ts, 0)]"
    |   "conflict_pos (Var x,Fun g ss,n) = [(Var x, Fun g ss,n)]"
    |   "conflict_pos (Fun f ts,Fun g ss,n) = 
            (if f = g \<and> length ts = length ss 
              then concat (map (\<lambda> (s,t). conflict_pos (s,t,n)) (zip ts ss))
              else [(Fun f ts, Fun g ss,n)])" 
declare size_simps[simp del]

lemma term_eq_via_conflict_pos_empty: 
   "(s = t \<cdot> \<mu> ^^ n) = (set (conflict_pos (s,t,n)) = {})"
proof (induct rule: conflict_pos.induct[of "\<lambda> (s,t,n). (s = t \<cdot> \<mu> ^^ n) = (set (conflict_pos (s,t,n)) = {})" "(s,t,n)", unfolded split])
  case (5 f ss g ts n)
  show ?case
  proof (cases "(f,length ss) = (g,length ts)")
    case False
    thus ?thesis by auto
  next
    case True
    hence f: "f = g" and l: "length ss = length ts" by auto
    hence l': "length ss = length (map (\<lambda> t. t \<cdot> \<mu> ^^ n) ts)" by auto
    have "(Fun f ss = (Fun g ts) \<cdot> \<mu> ^^ n) = (ss = map (\<lambda> t. t \<cdot> \<mu> ^^ n) ts)" unfolding f by auto
    also have "... = (\<forall> i < length ss. ss ! i = ts ! i \<cdot> \<mu> ^^ n)"
      using nth_equalityI[OF l'] l by auto
    also have "... = (\<forall> i < length ss. set (conflict_pos (ss ! i,ts ! i,n)) = {})"
    proof -
      {
        fix i
        assume i: "i < length ss"
        have "(ss ! i = ts ! i \<cdot> \<mu> ^^ n) = (set (conflict_pos (ss ! i,ts ! i,n)) = {})"
          by (rule 5[OF _ _ refl], insert f l i, auto simp: set_zip)
      }
      thus ?thesis by blast
    qed
    also have "... = (set (conflict_pos (Fun f ss,Fun g ts,n)) = {})"
      unfolding conflict_pos.simps 
      unfolding f l 
      unfolding zip_nth_conv[OF l] by auto
    finally
    show ?thesis .
  qed
qed auto

lemma conflict_pos_fun: assumes "(f,length ss) = (g,length ts)"
  shows "set (conflict_pos (Fun f ss, Fun g ts,n)) = \<Union> {set (conflict_pos (ss ! i, ts ! i, n)) | i. i < length ts}"
  using assms
  by (force simp: set_zip)


lemma term_subst_eq_via_conflict_pos: 
   "(s \<cdot> \<sigma> = t \<cdot> \<mu> ^^ n \<cdot> \<sigma>) = (\<forall> u v m. (u,v,m) \<in> set (conflict_pos (s,t,n)) \<longrightarrow> (u \<cdot> \<sigma> = v \<cdot> \<mu> ^^ m \<cdot> \<sigma>))"
proof (induct rule: conflict_pos.induct[of "\<lambda> (s,t,n). (s \<cdot> \<sigma> = t \<cdot> \<mu> ^^ n \<cdot> \<sigma>) = ((\<forall> u v m. (u,v,m) \<in> set (conflict_pos (s,t,n)) \<longrightarrow> (u \<cdot> \<sigma> = v \<cdot> \<mu> ^^ m \<cdot> \<sigma>)))" "(s,t,n)", unfolded split])
  case (2 x y)
  show ?case by (cases "x = y", auto)
next
  case (5 f ss g ts n)
  show ?case
  proof (cases "(f,length ss) = (g,length ts)")
    case False
    thus ?thesis by auto
  next
    case True
    hence f: "f = g" and l: "length ss = length ts" by auto
    let ?f = "\<lambda> t. t \<cdot> \<sigma>"
    let ?g = "\<lambda> t. t \<cdot> \<mu> ^^ n \<cdot> \<sigma>"
    have "(Fun f ss \<cdot> \<sigma> = (Fun g ts) \<cdot> \<mu> ^^ n \<cdot> \<sigma>) = (map (\<lambda> t. t \<cdot> \<sigma>) ss = map (\<lambda> t. t \<cdot> \<mu> ^^ n \<cdot> \<sigma>) ts)" (is "?l = _") unfolding f by auto
    also have "... = (\<forall> i < length ts. ss ! i \<cdot> \<sigma> = ts ! i \<cdot> \<mu> ^^ n \<cdot> \<sigma>)" (is "_ = ?r")
      using nth_map_conv[OF l, of ?f ?g] map_nth_conv[of ?f ss ?g ts] unfolding l by blast
    finally have id: "?l = ?r" .
    from l have id3: "set (zip ss ts) = {(ss ! i, ts ! i) | i. i < length ts}"
      unfolding set_zip by auto
    note 5 = 5[unfolded f l id3, OF conjI[OF refl refl] _ refl]
    from 5 have 5: "\<And> i. i < length ts \<Longrightarrow> (ss ! i \<cdot> \<sigma> = ts ! i \<cdot> \<mu> ^^ n \<cdot> \<sigma>) = (\<forall> u v m. (u,v,m) \<in> set (conflict_pos (ss ! i, ts ! i, n)) \<longrightarrow> u \<cdot> \<sigma> = v \<cdot> \<mu> ^^ m \<cdot> \<sigma>)" by blast
    let ?l = "\<forall> u v m. (u,v,m) \<in> \<Union> {set (conflict_pos (ss ! i, ts ! i, n)) | i. i < length ts} \<longrightarrow> u \<cdot> \<sigma> = v \<cdot> \<mu> ^^ m \<cdot> \<sigma>"
    show ?thesis unfolding id conflict_pos_fun[OF True] l 
    proof
      assume ?l
      show ?r
      proof (intro allI impI)
        fix i
        assume i: "i < length ts"
        show "ss ! i \<cdot> \<sigma> = ts ! i \<cdot> \<mu> ^^ n \<cdot> \<sigma>"
          unfolding 5[OF i] using `?l` i by blast
      qed
    next
      assume ?r
      show ?l
      proof (intro allI impI)
        fix u v m
        assume "(u,v,m) \<in> \<Union>{set (conflict_pos (ss ! i, ts ! i, n)) | i. i < length ts}"
        then obtain i where mem: "(u,v,m) \<in> set (conflict_pos (ss ! i, ts ! i, n))" and i: "i < length ts"
          by auto
        from 5[OF i] mem show "u \<cdot> \<sigma> = v \<cdot> \<mu> ^^ m \<cdot> \<sigma>" using `?r` i by blast
      qed
    qed
  qed
qed auto

definition conflict_terms :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term set"
  where "conflict_terms s t \<equiv> \<Union> (\<lambda> u. {v . u \<unrhd> v}) ` (subst_range \<mu> \<union> {s,t})"

lemma conflict_terms_subst: "conflict_terms s (\<mu> x) \<subseteq> conflict_terms s (Var x)"
  "conflict_terms (\<mu> x) s \<subseteq> conflict_terms (Var x) s"
proof -
  have "subst_range \<mu> \<union> {s,\<mu> x} \<subseteq> subst_range \<mu> \<union> {s,Var x}"
    by (cases "x \<in> subst_domain \<mu>", auto simp: subst_domain_def)
  thus "conflict_terms s (\<mu> x) \<subseteq> conflict_terms s (Var x)"
  "conflict_terms (\<mu> x) s \<subseteq> conflict_terms (Var x) s"
    unfolding conflict_terms_def by auto
qed 

lemma conflict_terms_refl: "s \<in> conflict_terms s t" "t \<in> conflict_terms s t"
  unfolding conflict_terms_def by auto

lemma conflict_terms_arg: assumes "i < length ss"
  shows "conflict_terms (ss ! i) t \<subseteq> conflict_terms (Fun f ss) t" "conflict_terms t (ss ! i) \<subseteq> conflict_terms t (Fun f ss)" 
proof -
  from assms have "ss ! i \<in> set ss" by auto 
  hence "{v. ss ! i \<unrhd> v} \<subseteq> {v. Fun f ss \<unrhd> v}" by auto
  thus "conflict_terms (ss ! i) t \<subseteq> conflict_terms (Fun f ss) t" "conflict_terms t (ss ! i) \<subseteq> conflict_terms t (Fun f ss)" 
    unfolding conflict_terms_def  
    unfolding image_Un by auto
qed

lemma conflict_terms_mono: assumes "{u,v} \<subseteq> conflict_terms s t"
  shows "conflict_terms u v \<subseteq> conflict_terms s t" (is "?c u v \<subseteq> _")
proof -
  note c = conflict_terms_def
  {
    fix u s
    assume "u \<in> ?c s s"
    hence "?c u u \<subseteq> ?c s s" unfolding c by (auto intro: supteq_trans)
  } note main = this
  {
    fix u s t
    assume "u \<in> ?c s t"
    hence "u \<in> ?c s s \<or> u \<in> ?c t t" unfolding c by auto
    with main[of u s] main[of u t] have "?c u u \<subseteq> ?c s s \<union> ?c t t" by blast
    hence "?c u u \<subseteq> ?c s t" unfolding c by blast
  } note main = this
  from assms have mem: "u \<in> ?c s t" "v \<in> ?c s t" by auto
  from main[OF mem(1)] main[OF mem(2)] have "?c u u \<union> ?c v v \<subseteq> ?c s t" by auto
  thus ?thesis unfolding c by auto
qed

lemma conflict_pos: 
  assumes "(u,v,m) \<in> set (conflict_pos (s,t,n))"
  shows "({u,v} \<subseteq> conflict_terms s t \<and> eroot u \<noteq> eroot v \<and> (is_Var v \<longrightarrow> is_Var u \<and> m = 0)) \<and> (\<exists> k p . n = m + k \<and> p \<in> poss s \<and> p \<in> poss (t \<cdot> \<mu> ^^ k) \<and> ((s |_ p, t \<cdot> \<mu> ^^ k |_ p) = (u,v) \<or> (s |_ p, t \<cdot> \<mu> ^^ k |_ p) = (v,u) \<and> m = 0))"
proof -
  let ?ct = "\<lambda> s t. conflict_terms s t"
  let ?cp = "\<lambda> s t n. set (conflict_pos (s,t,n))"
  let ?k = "\<lambda> s t n u v m k p. n = m + k \<and> p \<in> poss s \<and> p \<in> poss (t \<cdot> \<mu> ^^ k) \<and> ((s |_ p, t \<cdot> \<mu> ^^ k |_ p) = (u,v) \<or> (s |_ p, t \<cdot> \<mu> ^^ k |_ p) = (v,u) \<and> m = 0)"
  let ?p = "\<lambda> s t u v m. {u,v} \<subseteq> ?ct s t \<and> eroot u \<noteq> eroot v \<and> (is_Var v \<longrightarrow> is_Var u \<and> m = 0)"
  let ?P = "\<lambda> s t n u v m. ?p s t u v m \<and> (\<exists> k p. ?k s t n u v m k p)"
  let ?P' = "\<lambda> (s,t,n). (\<forall> u v m. (u,v,m) \<in> ?cp s t n \<longrightarrow> ?P s t n u v m)"
  show ?thesis
  proof (induct rule: conflict_pos.induct[of ?P' "(s,t,n)", unfolded split, THEN spec, THEN spec, THEN spec, THEN mp[OF _ assms]])
    case (1 s x n)
    let ?s = s let ?t = "Var x" let ?n = "Suc n"
    let ?s' = s let ?t' = "\<mu> x" let ?n' = "n"
    show ?case
    proof (intro allI impI)
      fix u v m
      assume "(u,v,m) \<in> ?cp ?s ?t ?n"
      hence "(u,v,m) \<in> ?cp ?s' ?t' ?n'" by simp
      from 1[THEN spec, THEN spec, THEN spec, THEN mp[OF _ this]]
      obtain k p where p: "?p ?s' ?t' u v m" and k: "?k ?s' ?t' ?n' u v m k p" by blast
      show "?P ?s ?t ?n u v m"
        by (rule conjI[OF _ exI[of _ "Suc k"]], rule conjI[OF subset_trans[OF _ conflict_terms_subst(1)]], insert p k, auto)
    qed
  next
    case (2 x y)
    let ?s = "Var x" let ?t = "Var y" let ?n = 0
    show ?case
    proof (intro allI impI)
      fix u v m
      assume mem: "(u,v,m) \<in> ?cp ?s ?t ?n"
      hence xy: "x \<noteq> y" by (cases "x = y", auto)
      with mem have m: "m = 0" and disj: "u = Var x \<and> v = Var y \<or> u = Var y \<and> v = Var x" by auto
      from disj have ct: "{u,v} \<subseteq> ?ct ?s ?t" using conflict_terms_refl by auto
      from disj ct xy have p: "?p ?s ?t u v m" unfolding m by auto
      show "?P ?s ?t ?n u v m"
        by (rule conjI[OF p], rule exI, rule exI[of _ \<epsilon>], insert disj m, auto)
    qed
  next
    case (3 f ts y)
    let ?s = "Fun f ts" let ?t = "Var y" let ?n = 0
    show ?case
    proof (intro allI impI)
      fix u v m
      assume mem: "(u,v,m) \<in> ?cp ?s ?t ?n"
      hence uvm: "u = ?t" "v = ?s" "m = ?n" by auto
      from uvm have ct: "{u,v} \<subseteq> ?ct ?s ?t" using conflict_terms_refl by auto
      from ct uvm have p: "?p ?s ?t u v m" by auto
      show "?P ?s ?t ?n u v m"
        by (rule conjI[OF p], rule exI, rule exI[of _ \<epsilon>], insert uvm, auto)
    qed
  next 
    case (4 x g ss n)
    let ?s = "Var x" let ?t = "Fun g ss" let ?n = n
    show ?case
    proof (intro allI impI)
      fix u v m
      assume mem: "(u,v,m) \<in> ?cp ?s ?t ?n"
      hence uvm: "u = ?s" "v = ?t" "m = ?n" by auto
      from uvm have ct: "{u,v} \<subseteq> ?ct ?s ?t" using conflict_terms_refl by auto
      from ct uvm have p: "?p ?s ?t u v m" by auto
      show "?P ?s ?t ?n u v m"
        by (rule conjI[OF p], rule exI, rule exI[of _ \<epsilon>], insert uvm, auto)
    qed
  next
    case (5 f ss g ts n)
    let ?s = "Fun f ss" let ?t = "Fun g ts" let ?n = n
    show ?case
    proof (intro allI impI)
      fix u v m
      assume mem: "(u,v,m) \<in> ?cp ?s ?t ?n"
      show "?P ?s ?t ?n u v m"
      proof (cases "(f,length ss) = (g,length ts)")
        case False
        with mem have uvm: "u = ?s" "v = ?t" "m = ?n" by auto
        from uvm have ct: "{u,v} \<subseteq> ?ct ?s ?t" using conflict_terms_refl by auto
        from ct uvm False have p: "?p ?s ?t u v m" by auto
        show "?P ?s ?t ?n u v m"
          by (rule conjI[OF p], rule exI, rule exI[of _ \<epsilon>], insert uvm, auto)
      next
        case True
        from conflict_pos_fun[OF this, of n] mem
        obtain i where i: "i < length ts" and mem: "(u,v,m) \<in> ?cp (ss ! i) (ts ! i) n" by auto
        have "?P (ss ! i) (ts ! i) n u v m"
          by (rule 5[OF _ _ refl, THEN spec, THEN spec, THEN spec, THEN mp[OF _ mem]], insert True i, auto simp: set_zip)
        then obtain k p where p: "?p (ss ! i) (ts ! i) u v m" and k: "?k (ss ! i) (ts ! i) n u v m k p" by auto
        let ?p = "i <# p"
        show ?thesis 
        proof (intro conjI exI)
          show "?p \<in> poss (?t \<cdot> \<mu> ^^ k)" using k i by auto
        next
          have "{u,v} \<subseteq> ?ct (ss ! i) (ts ! i)" using p by auto
          also have "... \<subseteq> ?ct ?s (ts ! i)" by (rule conflict_terms_arg, insert i True, auto)
          also have "... \<subseteq> ?ct ?s ?t" by (rule conflict_terms_arg[OF i])
          finally show "{u,v} \<subseteq> ?ct ?s ?t" .
        qed (insert k p i True, auto)
      qed
    qed
  qed
qed
end
  
locale fixed_finite_subst = fixed_subst \<mu>
  for \<mu> :: "('f,'v)subst" +
  assumes finite: "finite (subst_domain \<mu>)"
begin



definition nident_solution :: "('f,'v)nident_prob \<Rightarrow> nat \<Rightarrow> bool"
  where "nident_solution ip n \<equiv> case ip of (s,t,m) \<Rightarrow> s \<cdot> (\<mu> ^^ n) = t \<cdot> \<mu> ^^ m \<cdot> \<mu> ^^ n"

(* comments:
   - one can add conflict (x,y) with x,y \<notin> dom \<mu> 
     (but this will be detected one step later)
   - one can drop case (x,f(..)) with x \<notin> V_incr, but this may take a while
     until a duplicate is created, but for paper this is a nice result     
*)

function ident_solve' :: "('f,'v)nident_prob set \<Rightarrow> ('f,'v)nident_prob \<Rightarrow> nat option" 
  where
  "ident_solve' cps st = (let cp = set (conflict_pos st) in
        if (\<exists> f ss t n. (Fun f ss, t,n) \<in> cp) then None
        else if (\<exists> x f ss n. (Var x, Fun f ss,n) \<in> cp \<and> (x \<notin> V_incr \<or> x \<in> vars_term (Fun f ss))) then None
        else if (\<exists> x s t n m. (Var x, s,n) \<in> cp \<and> is_Fun s \<and> (Var x, t,m) \<in> cps \<and> is_Fun t \<and> mgu (s \<cdot> \<mu> ^^ n) (t \<cdot> \<mu> ^^ m) = None) then None
        else if (\<exists> u v n m. (u, v,n) \<in> cp \<and> (u,v,m) \<in> cps) then None
        else do {
           is \<leftarrow> mapM (\<lambda> (u,v,m). ident_solve' (insert (u,v,m) cps) (u \<cdot> \<mu>, v, Suc m)) (conflict_pos st);
           Some (max_list (map (\<lambda> i. Suc i) is))
        })"
  by pat_completeness auto


fun ident_solve'_nxt where "ident_solve'_nxt s t n cps 0 = (cps,(s,t,n))"
            | "ident_solve'_nxt s t n cps (Suc i) = (SOME tri. ident_solve'_rel tri (ident_solve'_nxt s t n cps i)  \<and> \<not> ident_solve'_dom tri)"


termination ident_solve'
proof (intro allI, rule ccontr)
  fix input
  assume "\<not> ident_solve'_dom input" 
  then obtain s t n cps where ndom: "\<not> ident_solve'_dom (cps, (s,t,n))" by (cases input, force+)
  {
    fix i
    have "(i > 0 \<longrightarrow> ident_solve'_rel (ident_solve'_nxt s t n cps i) (ident_solve'_nxt s t n cps (i - 1))) \<and> \<not> ident_solve'_dom (ident_solve'_nxt s t n cps i)"
    proof (induct i)
      case 0
      thus ?case using ndom by auto
    next
      case (Suc i)
      let ?p = "\<lambda> tri. ident_solve'_rel tri (ident_solve'_nxt s t n cps i) \<and> \<not> ident_solve'_dom tri"
      have id: "ident_solve'_nxt s t n cps (Suc i) = (SOME tri. ?p tri)" by auto
      show ?case
      proof (cases "\<exists> tri. ?p tri")
        case True
        from someI_ex[OF True]
        show ?thesis by auto
      next
        case False
        have "ident_solve'_dom (ident_solve'_nxt s t n cps i)"
          by (rule,clarify, insert False, blast)
        with Suc show ?thesis by simp
      qed
    qed
  } note chain = this
  {
    fix i
    have "ident_solve'_rel (ident_solve'_nxt s t n cps (Suc i)) (ident_solve'_nxt s t n cps i)" "\<not> ident_solve'_dom (ident_solve'_nxt s t n cps i)"
      using chain[of "Suc i"] chain[of i] by auto
  } then
  obtain f where f: "\<And> i. ident_solve'_rel (f (Suc i)) (f i)" 
    by blast
  obtain ss ts ns cp where id: "\<And> i. ss i = fst (snd (f i)) \<and> ts i = fst (snd (snd (f i))) \<and> ns i = snd (snd (snd (f i))) \<and> cp i = fst (f i)" by force+
  from f id have 
    rel: "\<And> i. ident_solve'_rel (cp (Suc i),(ss (Suc i),ts (Suc i), ns (Suc i))) (cp i,(ss i,ts i, ns i))" 
    by auto
  let ?cp = "\<lambda> i. set (conflict_pos (ss i,ts i, ns i))"
  let ?p = "\<lambda> i u v m.  (u,v,m) \<in> ?cp i \<and> \<not> (\<exists> f ss. u = Fun f ss) \<and> (\<forall> n. (u,v,n) \<notin> cp i) \<and> cp (Suc i) = insert (u,v,m) (cp i) \<and> ss (Suc i) = u \<cdot> \<mu> \<and> ts (Suc i) = v"
  {
    fix i    
    from rel[of i] have "\<exists> u v m. ?p i u v m"
    proof(rule, clarify)
      fix u m
      assume cp: "(u,ts (Suc i),m) \<in> ?cp i" and 
        ins: "cp (Suc i) = insert (u,ts (Suc i),m) (cp i)" and
        nofun: "\<not> (\<exists> f ss t n'. (Fun f ss,t,n') \<in> ?cp i)" and
        nodup: "\<not> (\<exists> u' v' n' m'. (u',v',n') \<in> ?cp i \<and> (u',v',m') \<in> cp i)" and
        id: "ss (Suc i) = u \<cdot> \<mu>" 
      show ?thesis
        by (intro exI conjI allI, insert cp nodup nofun, unfold ins id, auto)
    qed
  }
  hence "\<forall> i. \<exists> u v m. ?p i u v m" ..
  from choice[OF this] obtain u where "\<forall> i. \<exists> v m. ?p i (u i) v m" ..
  from choice[OF this] obtain v where "\<forall> i. \<exists> m. ?p i (u i) (v i) m" ..
  from choice[OF this] obtain m where uvm: "\<And> i. ?p i (u i) (v i) (m i)" by blast
  hence cp: "\<And> i. (u i, v i, m i) \<in> set (conflict_pos (ss i, ts i, ns i))" by auto
  let ?ct = "conflict_terms (ss 0) (ts 0)"
  {
    fix i
    have "{u i,v i} \<subseteq> ?ct"
    proof (induct i)
      case 0
      show ?case using conflict_pos[OF cp[of 0]] by simp
    next
      case (Suc i)
      from uvm[of i] obtain x where x: "u i = Var x" by (cases "u i", auto)
      from conflict_pos[OF cp[of "Suc i"]]  uvm[of i] have "{u (Suc i), v (Suc i)} \<subseteq> conflict_terms (\<mu> x) (v i)" using x by auto
      also have "... \<subseteq> conflict_terms (u i) (v i)" unfolding x by (rule conflict_terms_subst)
      also have "... \<subseteq> conflict_terms (ss 0) (ts 0)" using conflict_terms_mono[OF Suc(1)] .
      finally show ?case .
    qed
  } note ct = this
  {
    fix i
    have "cp i = cp 0 \<union> (\<lambda> i. (u i,v i,m i)) ` {0 ..< i}"
    proof (induct i)
      case 0 show ?case by simp
    next
      case (Suc i)
      have id: "cp (Suc i) = insert (u i,v i,m i) (cp i)" using uvm[of i] by simp
      show ?case unfolding id Suc atLeastLessThanSuc by auto
    qed
  } note cp = this
  {
    fix i j :: nat
    assume "i < j"
    with cp[of j] have "(u i, v i, m i) \<in> cp j" by auto
    with uvm[of j] have "(u i, v i) \<noteq> (u j, v j)" by auto
  } note neq = this
  have finite: "finite ?ct" unfolding conflict_terms_def subst_range.simps
  proof (rule, rule, rule, rule, rule, rule finite, rule, rule, rule)
    fix M
    assume "M \<in> (\<lambda> u. {v. u \<unrhd> v}) ` (\<mu> ` subst_domain \<mu> \<union> {ss 0, ts 0})"
    then obtain u where M: "M = {v. u \<unrhd> v}" by auto
    show "finite M" unfolding M by (rule finite_subterms)
  qed
  have finite: "finite (?ct \<times> ?ct)"
    by (rule, rule finite, rule finite)
  let ?uv = "\<lambda> i. (u i, v i)"
  have finite: "finite (range ?uv)" 
    by (rule finite_subset[OF _ finite], insert ct, auto)
  from pigeonhole_infinite[OF _ finite] obtain uv where inf: "infinite {i. ?uv i = uv}" by auto 
  note inf = inf[unfolded infinite_nat_iff_unbounded, THEN spec]
  from inf[of 0] obtain n where n: "(u n, v n) = uv" by auto
  from inf[of n] obtain m where m: "(u m, v m) = uv" "m > n" by auto
  from neq[OF m(2)] m n show False by simp
qed


lemma ident_solve'_sound: 
  shows "ident_solve' cps st = Some i \<longrightarrow> nident_solution st i"
proof (induct cps st arbitrary: i rule: ident_solve'.induct)
  case (1 cps st i)
  obtain s t n where st: "st = (s,t,n)" by (cases st, force)
  note 1 = 1[unfolded st]
  show ?case unfolding st
  proof (intro impI)
    assume sol: "ident_solve' cps (s,t,n) = Some i"
    note simp = ident_solve'.simps[of "cps" st, unfolded Let_def sol st]
    let ?cp = "set (conflict_pos (s,t,n))"
    let ?c1 = "\<exists> f ss t n. (Fun f ss, t, n) \<in> ?cp"
    from simp have c1: "?c1 = False" by (cases ?c1, auto)
    note simp = simp[unfolded this]
    let ?c2 = "\<exists> x f ss n. (Var x, Fun f ss,n) \<in> ?cp \<and> (x \<notin> V_incr \<or> x \<in> vars_term (Fun f ss))"
    from simp have c2: "?c2 = False" by (cases ?c2, auto)
    note simp = simp[unfolded this]
    let ?c4 = "\<exists> x s t n m. (Var x, s,n ) \<in> ?cp \<and> is_Fun s \<and> (Var x,t,m) \<in> cps \<and> is_Fun t \<and> mgu (s \<cdot> \<mu> ^^ n) (t \<cdot> \<mu> ^^ m) = None"
    from simp have c4: "?c4 = False" by (cases ?c4, auto)
    note simp = simp[unfolded this]
    let ?c5 = "\<exists> u v n m. (u,v,n) \<in> ?cp \<and> (u, v,m) \<in> cps"
    from simp have c5: "?c5 = False" by (cases ?c5, auto)
    note simp = simp[unfolded this]
    let ?recs = "mapM (\<lambda>(u,v,m). ident_solve' (insert (u,v,m) cps) (u \<cdot> \<mu>, v, Suc m)) (conflict_pos (s,t,n))"
    obtain iis where recs: "?recs = Some iis" using simp by (cases ?recs, auto)
    note simp = simp[unfolded this]
    let ?res = "map (\<lambda> i. Suc i) iis"
    have i: "i = max_list ?res" using simp by simp
    show "nident_solution (s,t,n) i" unfolding nident_solution_def split
      term_subst_eq_via_conflict_pos[of s]
    proof (intro allI impI)
      fix u v m
      assume uv: "(u,v,m) \<in> ?cp"
      then obtain idx where idx: "idx < length (conflict_pos (s,t,n))" and uvi: "conflict_pos (s,t,n) ! idx = (u,v,m)" 
        unfolding set_conv_nth by auto
      from mapM_Some_idx[OF recs idx, unfolded uvi split] obtain j where 
        sol: "ident_solve' (insert (u,v,m) cps) (u \<cdot> \<mu>, v, Suc m) = Some j" and j: "j = iis ! idx" by auto
      from 1[OF refl, THEN mp[OF _ sol], OF _ _ _ _ uv] c1 c2 c4 c5
      have "nident_solution (u \<cdot> \<mu>, v, Suc m) j" by auto
      from this[unfolded nident_solution_def split]
      have "u \<cdot> \<mu> ^^ Suc j = v \<cdot> \<mu> ^^ (Suc m + j)"  unfolding subst_power_compose_distrib by simp
      also have "... = v \<cdot> \<mu> ^^ (m + Suc j)" by auto
      also have "... = v \<cdot> \<mu> ^^ m \<cdot> \<mu> ^^ Suc j" unfolding subst_power_compose_distrib by simp
      finally have sol: "u \<cdot> \<mu> ^^ Suc j = v \<cdot> \<mu> ^^ m \<cdot> \<mu> ^^ Suc j" .        
      from idx j have "Suc j \<in> set ?res" unfolding mapM_Some[OF recs, THEN conjunct1] by auto
      from max_list[OF this] have "Suc j \<le> i" unfolding i .
      hence "i = Suc j + (i - Suc j)" by auto
      then obtain k where i: "i = Suc j + k" ..
      from arg_cong[OF sol, of "\<lambda> u. u \<cdot> \<mu> ^^ k"]
      show "u \<cdot> \<mu> ^^ i = v \<cdot> \<mu> ^^ m \<cdot> \<mu> ^^ i" unfolding i subst_power_compose_distrib by simp
    qed
  qed
qed


lemma ident_solution_subt: assumes sol: "ident_solution (S,T) m"
  and p: "p \<in> poss (S \<cdot> \<mu> ^^ n)" "p \<in> poss (T \<cdot> \<mu> ^^ n)"
  shows "S \<cdot> \<mu> ^^ n |_ p \<cdot> \<mu> ^^ m = T \<cdot> \<mu> ^^ n |_ p \<cdot> \<mu> ^^ m"
proof -
  let ?sol = "\<lambda> s t n. s \<cdot> \<mu> ^^ n = t \<cdot> \<mu> ^^ n"
  from sol have sol: "?sol S T m" unfolding ident_solution_def by auto
  from arg_cong[OF sol, of "\<lambda> s. s \<cdot> \<mu> ^^ n"]
  have "?sol S T (m + n)" unfolding subst_power_compose_distrib by auto
  hence "?sol S T (n + m)" unfolding add_commute[of m] .
  hence sol: "?sol (S \<cdot> \<mu> ^^ n) (T \<cdot> \<mu> ^^ n) m" unfolding subst_power_compose_distrib by simp
  show ?thesis unfolding 
    subst_subt_at_distr[OF p(1), of "\<mu> ^^ m"]
    subst_subt_at_distr[OF p(2), of "\<mu> ^^ m"]
    sol ..
qed


lemma ident_solve'_complete: assumes
  SOL: "ident_solution (S,T) M"
  shows "\<forall> p n. p \<in> poss (S \<cdot> \<mu> ^^ n) 
      \<longrightarrow>  p \<in> poss (T \<cdot> \<mu> ^^ n)
      \<longrightarrow> ((S \<cdot> \<mu> ^^ n) |_ p, (T \<cdot> \<mu> ^^ n) |_ p) \<in> {(s,t \<cdot> \<mu> ^^ N), (t \<cdot> \<mu> ^^ N, s)}
      \<longrightarrow> (\<forall> u v m. (u,v,m) \<in> cps \<longrightarrow> (\<exists> p' n'. (m > 0 \<longrightarrow> is_Fun v) \<and> p' \<le> p \<and> n' < n \<and> p' \<in> poss (S \<cdot> \<mu> ^^ n') \<and> p' \<in> poss (T \<cdot> \<mu> ^^ n') \<and> eroot u \<noteq> eroot v \<and> ((S \<cdot> \<mu> ^^ n') |_ p', (T \<cdot> \<mu> ^^ n') |_ p') \<in> {(u,v \<cdot> \<mu> ^^ m),(v \<cdot> \<mu> ^^ m,u)}))
      \<longrightarrow> (\<exists> i. ident_solve' cps (s,t,N) = Some i)" (is "?P s t N cps")
proof (induct rule: ident_solve'.induct[of "\<lambda> cps (s,t,N). ?P s t N cps" cps "(s,t,N)", unfolded split])
  case (1 cps st)
  obtain s t N where st: "st = (s,t,N)" by (cases st, force)
  note 1 = 1[unfolded st split]
  let ?sol = "\<lambda> s t N n. s \<cdot> \<mu> ^^ n = t \<cdot> \<mu> ^^ N \<cdot> \<mu> ^^ n"
  let ?cp = "set (conflict_pos (s,t,N))"
  let ?P = "\<lambda> u' v' m p'' n' P N. (m > 0 \<longrightarrow> is_Fun v') \<and> p'' \<le> P \<and> n' < N \<and> p'' \<in> poss (S \<cdot> \<mu> ^^ n') \<and> p'' \<in> poss (T \<cdot> \<mu> ^^ n') \<and> eroot u' \<noteq> eroot v' \<and> (S \<cdot> \<mu> ^^ n' |_ p'', T \<cdot> \<mu> ^^ n' |_ p'') \<in> {(u',v' \<cdot> \<mu> ^^ m), (v' \<cdot> \<mu> ^^ m, u')}"
  let ?Q = "\<lambda> u' v' m P N. \<exists> p'' n'. ?P u' v' m p'' n' P N"
  note SOL = ident_solution_subt[OF SOL]
  show ?case unfolding st split
  proof (intro allI impI)
    fix p n
    assume p: "p \<in> poss (S \<cdot> \<mu> ^^ n)" "p \<in> poss (T \<cdot> \<mu> ^^ n)"
      and st: "(S \<cdot> \<mu> ^^ n |_ p, T \<cdot> \<mu> ^^ n |_ p) \<in> {(s,t \<cdot> \<mu> ^^ N),(t \<cdot> \<mu> ^^ N,s)}"
      and CPS: "\<forall> u v m. (u,v,m) \<in> cps \<longrightarrow> ?Q u v m p n"
    from SOL[OF p] st have sol: "?sol s t N M" by auto
    {
      fix u v m
      assume "(u,v,m) \<in> cps"
      from CPS[THEN spec, THEN spec, THEN spec, THEN mp[OF _ this]]
      obtain p' n' where p': "p' \<in> poss (S \<cdot> \<mu> ^^ n')" "p' \<in> poss (T \<cdot> \<mu> ^^ n')"
        and id: "(S \<cdot> \<mu> ^^ n' |_ p', T \<cdot> \<mu> ^^ n' |_ p') \<in> {(u,v \<cdot> \<mu> ^^ m),(v \<cdot> \<mu> ^^ m, u)}" by blast
      from SOL[OF p'] id have "?sol u v m M" by auto
    } note sol_cps = this
    note simp = ident_solve'.simps[of cps "(s,t,N)", unfolded Let_def]
    let ?c1 = "\<exists> f ss t n. (Fun f ss, t, n) \<in> ?cp"
    let ?c2 = "\<exists> x f ss n. (Var x, Fun f ss,n) \<in> ?cp \<and> (x \<notin> V_incr \<or> x \<in> vars_term (Fun f ss))"
    let ?c4 = "\<exists> x s t n m. (Var x, s,n ) \<in> ?cp \<and> is_Fun s \<and> (Var x,t,m) \<in> cps \<and> is_Fun t \<and> mgu (s \<cdot> \<mu> ^^ n) (t \<cdot> \<mu> ^^ m) = None"
    let ?c5 = "\<exists> u v n m. (u,v,n) \<in> ?cp \<and> (u, v,m) \<in> cps"
    from sol[unfolded term_subst_eq_via_conflict_pos[of s]]
    have cps: "\<And> u v m. (u,v,m) \<in> ?cp \<Longrightarrow> ?sol u v m M" by auto
    {
      assume ?c1
      then obtain f g ss t n where cp: "(Fun f ss, t,n) \<in> ?cp" by auto
      from arg_cong[OF cps[OF this], of eroot] conflict_pos[OF cp, THEN conjunct1] have False by auto
    }
    hence c1: "?c1 = False" by blast
    note simp = simp[unfolded c1]
    {
      assume ?c2
      then obtain x f ss n where mem: "(Var x, Fun f ss,n) \<in> ?cp" and disj: "x \<notin> V_incr \<or> x \<in> vars_term (Fun f ss)" by blast+
      from cps[OF mem] have id: "Var x \<cdot> \<mu> ^^ M = Fun f ss \<cdot> \<mu> ^^ n \<cdot> \<mu> ^^ M" by blast
      from disj have False
      proof 
        assume x: "x \<notin> V_incr"
        from x[unfolded V_incr_def] have "is_Var (Var x \<cdot> \<mu> ^^ M)" by auto
        with id show False by (cases "Var x \<cdot> \<mu> ^^ M", auto)
      next
        assume "x \<in> vars_term (Fun f ss)"
        hence "Fun f ss \<rhd> Var x" by auto
        hence large: "Fun f ss \<cdot> \<mu> ^^ M \<rhd> Var x \<cdot> \<mu> ^^ M" by (rule supt_stable)
        hence "size (Var x \<cdot> \<mu> ^^ M) < size (Fun f ss \<cdot> \<mu> ^^ M)" by (rule supt_size)
        also have "... \<le> size (Fun f ss \<cdot> \<mu> ^^ M \<cdot> \<mu> ^^ n)" by (rule size_subst)
        also have "... = size (Fun f ss \<cdot> \<mu> ^^ (n + M))" unfolding add_commute[of n]  subst_power_compose_distrib by auto
        also have "... = size (Var x \<cdot> \<mu> ^^ M)" unfolding subst_power_compose_distrib id by simp
        finally show False by simp
      qed
    }
    hence c2: "?c2 = False" by auto
    note simp = simp[unfolded c2]
    {
      assume ?c4
      then obtain x u v n m where cp: "(Var x, u,n) \<in> ?cp" "(Var x, v,m) \<in> cps" and mgu: "mgu (u \<cdot> \<mu> ^^ n) (v \<cdot> \<mu> ^^ m) = None" by blast
      from cps[OF cp(1)] sol_cps[OF cp(2)] have "u \<cdot> \<mu> ^^ n \<cdot> \<mu> ^^ M = v \<cdot> \<mu> ^^ m \<cdot> \<mu> ^^ M" by simp
      with mgu_complete[OF mgu, unfolded unifiers_def] have False  by auto
    }
    hence c4: "?c4 = False" by blast
    note simp = simp[unfolded c4]
    {
      assume ?c5
      then obtain u v n1 n2 where cp: "(u,v,n1) \<in> ?cp" "(u,v,n2) \<in> cps" by auto
      from conflict_pos[OF cp(1)] obtain k1 q where 
        n1: "N = n1 + k1" and
        var: "is_Var v \<longrightarrow> n1 = 0" and
        eroot: "eroot u \<noteq> eroot v"  and
        q: "q \<in> poss s" "q \<in> poss (t \<cdot> \<mu> ^^ k1)"  and
        uv: "(u = s |_ q \<and> v = (t \<cdot> \<mu> ^^ k1 |_ q)) \<or> (v = s |_ q \<and> u = (t \<cdot> \<mu> ^^ k1 |_ q) \<and> n1 = 0)" by auto
      from CPS[THEN spec, THEN spec, THEN spec, THEN mp[OF _ cp(2)]]
      obtain p' n' where P: "?P u v n2 p' n' p n" by auto
      hence "n' < n" by auto
      hence  "\<exists> k. k > 0 \<and> n = n' + k" by arith
      then obtain k where k: "k > 0" and n: "n = n' + k" by auto
      from P have "p' \<le> p" by auto
      then obtain q' where pp': "p = p' <#> q'" unfolding less_eq_pos_def by auto
      let ?\<sigma> = "\<mu> ^^ k"
      let ?p = "q' <#> q"
      from P have p': "p' \<in> poss (S \<cdot> \<mu> ^^ n')" "p' \<in> poss (T \<cdot> \<mu> ^^ n')" by simp_all
      have p'1: "p' \<in> poss (S \<cdot> \<mu> ^^ n' \<cdot> ?\<sigma>)" using poss_imp_subst_poss[OF p'(1)] .
      have p'2: "p' \<in> poss (T \<cdot> \<mu> ^^ n' \<cdot> ?\<sigma>)" using poss_imp_subst_poss[OF p'(2)] .
      from P have uv': "(u,v \<cdot> \<mu> ^^ n2) \<in> {((S \<cdot> \<mu> ^^ n' |_ p'), T \<cdot> \<mu> ^^ n' |_ p')}\<^sup>\<leftrightarrow>" by auto
      hence "(u \<cdot> ?\<sigma> |_ ?p, v \<cdot> \<mu> ^^ n2 \<cdot> ?\<sigma> |_ ?p) \<in> {((S \<cdot> \<mu> ^^ n' |_ p' \<cdot> ?\<sigma>) |_ ?p, (T \<cdot> \<mu> ^^ n' |_ p' \<cdot> ?\<sigma>) |_ ?p)}\<^sup>\<leftrightarrow>" 
        (is "?pair \<in> ?set")
        by auto 
      have "?set = {(S \<cdot> \<mu> ^^ n' \<cdot> ?\<sigma> |_ p' <#> ?p, T \<cdot> \<mu> ^^ n' \<cdot> ?\<sigma> |_ p' <#> ?p)}\<^sup>\<leftrightarrow>"
        unfolding subt_at_subst[OF p'(1), symmetric] subt_at_subst[OF p'(2), symmetric]
        unfolding subt_at_append[OF p'1] subt_at_append[OF p'2] ..
      also have "... = {(S \<cdot> \<mu> ^^ n |_ p' <#> ?p, T \<cdot> \<mu> ^^ n |_ p' <#> ?p)}\<^sup>\<leftrightarrow>"
        unfolding n subst_power_compose_distrib by simp
      also have "... = {(S \<cdot> \<mu> ^^ n |_ p <#> q, T \<cdot> \<mu> ^^ n |_ p <#> q)}\<^sup>\<leftrightarrow>" unfolding pp' by simp
      also have "... = {(S \<cdot> \<mu> ^^ n |_ p |_ q, T \<cdot> \<mu> ^^ n |_ p |_ q)}\<^sup>\<leftrightarrow>" 
        unfolding subt_at_append[OF p(1)]  subt_at_append[OF p(2)] ..
      also have "... = {(u,v \<cdot> \<mu> ^^ n1)}\<^sup>\<leftrightarrow>" using uv 
        st[unfolded n1 add_commute[of n1] subst_power_compose_distrib subst_compose_apply_term_distrib]  
        by (auto simp: subt_at_subst[OF q(2)])
      finally have mem: "?pair \<in> {(u,v \<cdot> \<mu> ^^ n1)}\<^sup>\<leftrightarrow>" using `?pair \<in> ?set` by simp
      obtain pp \<sigma> v' where pp: "pp = ?p" and \<sigma>: "\<sigma> = ?\<sigma>" and v': "v' = v \<cdot> \<mu> ^^ n2" by auto
      from mem have mem: "(u \<cdot> \<sigma> |_ pp, v' \<cdot> \<sigma> |_ pp) \<in> {(u,v \<cdot> \<mu> ^^ n1)}\<^sup>\<leftrightarrow>" unfolding pp \<sigma> v' by auto
      (* now consider positions *)
      have "(pp \<in> poss (u \<cdot> \<sigma>) \<inter> poss (v' \<cdot> \<sigma>)) = (pp \<in> poss (S \<cdot> \<mu> ^^ n' \<cdot> \<sigma> |_p') \<inter> poss (T \<cdot> \<mu> ^^ n' \<cdot> \<sigma> |_ p'))" unfolding subt_at_subst[OF p'(1)] subt_at_subst[OF p'(2)] pp \<sigma> v' 
        using  uv' by auto
      also have "... = (pp \<in> poss (S \<cdot> \<mu> ^^ n |_ p') \<inter> poss (T \<cdot> \<mu> ^^ n |_ p'))" unfolding n subst_power_compose_distrib  \<sigma> by simp
      also have "..."
      proof -
        have "S \<cdot> \<mu> ^^ n |_ p' <#> q' = S \<cdot> \<mu> ^^ n |_ p' |_ q'" "T \<cdot> \<mu> ^^ n |_ p' |_ q' = T \<cdot> \<mu> ^^ n |_ p' <#> q'" unfolding 
          subt_at_append[OF p(1)[unfolded pp' poss_append_poss, THEN conjunct1]]
          subt_at_append[OF p(2)[unfolded pp' poss_append_poss, THEN conjunct1]] by auto
        hence one: "q \<in> poss (S \<cdot> \<mu> ^^ n |_ p' |_ q')" "q \<in> poss (T \<cdot> \<mu> ^^ n |_ p' |_ q')" using q st 
          unfolding pp' n1 add_commute[of n1] subst_power_compose_distrib by auto
        show ?thesis using 
          pos_append_poss[OF p(1)[unfolded pp' poss_append_poss, THEN conjunct2] one(1)] 
          pos_append_poss[OF p(2)[unfolded pp' poss_append_poss, THEN conjunct2] one(2)] unfolding pp
          by auto
      qed
      finally have pos: "pp \<in> poss (u \<cdot> \<sigma>)" "pp \<in> poss (v' \<cdot> \<sigma>)" by auto
      have False
      proof (cases "n1 = n2")
        case True
        let ?\<tau> = "\<sigma> \<circ>s \<sigma>" let ?q = "pp <#> pp" 
        from mem[unfolded True v'[symmetric]]
        have "(u \<cdot> \<sigma> |_ pp) = u \<and> (v' \<cdot> \<sigma> |_ pp) = v' \<or> (u \<cdot> \<sigma> |_ pp) = v' \<and> (v' \<cdot> \<sigma> |_ pp) = u" by auto
        hence "(u \<cdot> ?\<tau> |_ ?q) = u \<and> (v' \<cdot> ?\<tau> |_ ?q) = v' \<and> ?q \<in> poss (u \<cdot> ?\<tau>) \<and> ?q \<in> poss (v' \<cdot> ?\<tau>)"
        proof
          assume "(u \<cdot> \<sigma> |_ pp) = u \<and> (v' \<cdot> \<sigma> |_ pp) = v'"
          hence id: "(u \<cdot> \<sigma> |_ pp) = u" "(v' \<cdot> \<sigma> |_ pp) = v'" by auto
          note it = arg_cong[where f = "\<lambda> t. t \<cdot> \<sigma> |_ pp"]
          show ?thesis
            unfolding poss_append_poss
            unfolding subst_compose_apply_term_distrib 
            unfolding subt_at_append[OF poss_imp_subst_poss[OF pos(1)]] subt_at_subst[OF pos(1)]
            unfolding subt_at_append[OF poss_imp_subst_poss[OF pos(2)]] subt_at_subst[OF pos(2)]
            unfolding it[OF id(1)] it[OF id(2)] 
            unfolding id 
            using pos by auto
        next
          assume "(u \<cdot> \<sigma> |_ pp) = v' \<and> (v' \<cdot> \<sigma> |_ pp) = u"
          hence id: "(u \<cdot> \<sigma> |_ pp) = v'" "(v' \<cdot> \<sigma> |_ pp) = u" by auto
          note it = arg_cong[where f = "\<lambda> t. t \<cdot> \<sigma> |_ pp"]
          show ?thesis
            unfolding poss_append_poss
            unfolding subst_compose_apply_term_distrib 
            unfolding subt_at_append[OF poss_imp_subst_poss[OF pos(1)]] subt_at_subst[OF pos(1)]
            unfolding subt_at_append[OF poss_imp_subst_poss[OF pos(2)]] subt_at_subst[OF pos(2)]
            unfolding it[OF id(1)] it[OF id(2)] 
            unfolding id 
            using pos by auto
        qed
        hence urec: "u \<cdot> ?\<tau> |_ ?q = u" and upos: "?q \<in> poss (u \<cdot> ?\<tau>)" and
              vrec: "v' \<cdot> ?\<tau> |_ ?q = v'" and vpos: "?q \<in> poss (v' \<cdot> ?\<tau>)" by auto
        note uiter = iterate_term[OF urec upos, THEN conjunct1, unfolded subst_pow_mult]
        note viter = iterate_term[OF vrec vpos, THEN conjunct1, unfolded subst_pow_mult]
        let ?it = "\<lambda> u n :: nat. u \<cdot> \<mu> ^^ ((k + k) * n) |_ ?q ^ n"
        from eroot var have "u \<noteq> v'" unfolding v' True by (cases u, (cases v, auto)+)
        hence neq: "\<And> n. ?it u n \<noteq> ?it v' n" using uiter viter unfolding \<sigma> subst_power_compose_distrib[symmetric] 
          unfolding subst_pow_mult by auto
        from sol_cps[OF cp(2)] have "?sol u v n2 M" by simp
        hence "?sol u v' 0 M" unfolding v' by simp
        from arg_cong[OF this, of "\<lambda> s. s \<cdot> \<mu> ^^ ((k + k - 1)* M)"] 
        have "?sol u v' 0 (M + ( k + k - 1) * M)" unfolding subst_power_compose_distrib by simp
        with k have "?sol u v' 0 ((k + k) * M)" by (cases k, auto)
        with neq[of M] show False by auto
      next
        case False
        hence "n1 > 0 \<or> n2 > 0" by auto        
        with P var have nvar: "is_Fun v" by auto
        with cp(1) c1 obtain x where u: "u = Var x" by (cases u, simp, cases v, auto)
        from sol_cps[OF cp(2)] cps[OF cp(1)] have id: "v \<cdot> \<mu> ^^ (M + n2) = v \<cdot> \<mu> ^^ (M + n1)" 
          unfolding add_commute[of M] subst_power_compose_distrib by simp
        let ?n1 = "min n1 n2" let ?n2 = "max n1 n2 - ?n1"
        have id: "v \<cdot> \<mu> ^^ (M + ?n1) \<cdot> \<mu> ^^ ?n2 = v \<cdot> \<mu> ^^ (M + ?n1) \<and> ?n2 > 0" (is "?id \<and> _")
        proof (cases "n1 < n2")
          case True
          hence id1: "?n1 = n1" and id2: "?n2 = n2 - n1" by auto
          show ?thesis using id unfolding id2 unfolding id1 using True
            unfolding subst_compose_apply_term_distrib[symmetric] subst_power_compose_distrib[symmetric] by simp
        next
          case False with `n1 \<noteq> n2` 
          have True: "n2 < n1" by simp
          hence id1: "?n1 = n2" and id2: "?n2 = n1 - n2" by auto
          show ?thesis using id unfolding id2 unfolding id1 using True
            unfolding subst_compose_apply_term_distrib[symmetric] subst_power_compose_distrib[symmetric] by simp
        qed          
        hence ?id "?n2 > 0" by auto
        then obtain n3 where n3: "?n2 = Suc n3" by (cases ?n2, auto)
        from iterate_term[of _ _ \<epsilon>, simplified, OF `?id`]
        have id: "\<And> k. v \<cdot> \<mu> ^^ (M + ?n1 + k + n3 * k) = v \<cdot> \<mu> ^^ (M + ?n1)" unfolding n3
          unfolding subst_pow_mult subst_compose_apply_term_distrib[symmetric] subst_power_compose_distrib[symmetric] 
          by (simp add: field_simps)
        from mem
        have "(u \<cdot> \<sigma> |_ pp) = u \<and> (v' \<cdot> \<sigma> |_ pp) = v \<cdot> \<mu> ^^ n1 \<or> (u \<cdot> \<sigma> |_ pp) = v \<cdot> \<mu> ^^ n1 \<and> (v' \<cdot> \<sigma> |_ pp) = u" 
          (is "?one \<or> ?two") by auto
        thus False
        proof
          assume ?two
          hence v\<sigma>: "v' \<cdot> \<sigma> |_ pp = u" "u \<cdot> \<sigma> |_ pp = v \<cdot> \<mu> ^^ n1" by auto
          from nvar have nvar: "is_Fun (v' \<cdot> \<sigma>)" unfolding v' by auto
          from subt_at_imp_supteq[OF pos(2)] have "v' \<cdot> \<sigma> \<unrhd> u" unfolding v\<sigma> .
          moreover have "v' \<cdot> \<sigma> \<noteq> u" unfolding u using nvar by auto
          ultimately have supt: "v \<cdot> \<mu> ^^ n2 \<cdot> \<sigma> \<rhd> u" unfolding v' by auto
          from supt_stable[OF supt] have supt: "v \<cdot> \<mu> ^^ n2 \<cdot> \<sigma> \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>" .
          from subt_at_imp_supteq[OF pos(1)] have supteq: "u \<cdot> \<sigma> \<unrhd> v \<cdot> \<mu> ^^ n1" unfolding v\<sigma> .
          from supt_supteq_trans[OF supt supteq] have supt: "v \<cdot> \<mu> ^^ n2 \<cdot> \<sigma> \<cdot> \<sigma> \<rhd> v \<cdot> \<mu> ^^ n1" .
          {
            fix i j
            let ?v = "v \<cdot> \<mu> ^^ (n2 + k + k + i)"
            let ?v' = "v \<cdot> \<mu> ^^ (n2 + k + k + i + j)"
            let ?w = "v \<cdot> \<mu> ^^ (n1 + i)"
            from size_subst[of ?v "\<mu> ^^ j"]  have ge: "size (?v') \<ge> size ?v"
              unfolding subst_power_compose_distrib by simp
            from supt_size[OF supt_stable[OF supt]] have gt: "size (v \<cdot> \<mu> ^^ (n2 + k + k + i)) > size ?w" 
              unfolding \<sigma> 
              unfolding supt_supteq_conv 
              unfolding subst_power_compose_distrib by auto
            from ge gt have "?v' \<noteq> ?w" by auto
          } note neq = this
          from arg_cong[OF id[of "n2 + k + k"], of "\<lambda> t. t \<cdot> \<mu> ^^ n1"] neq[of "M + min n1 n2" "n3 * (n2 + k + k) + n1"]
          show False unfolding subst_compose_apply_term_distrib[symmetric] 
            subst_power_compose_distrib[symmetric] by (auto simp: field_simps)
        next
          assume ?one
          hence v\<sigma>: "v' \<cdot> \<sigma> |_ pp = v \<cdot> \<mu> ^^ n1" "u \<cdot> \<sigma> |_ pp = u" by auto
          from subt_at_imp_supteq[OF pos(1)] have "u \<cdot> \<sigma> \<unrhd> u" unfolding v\<sigma> by auto
          moreover have "u \<cdot> \<sigma> \<noteq> u" 
          proof
            assume "u \<cdot> \<sigma> = u" 
            hence "u \<cdot> \<sigma> |_ \<epsilon> = u" by auto
            from iterate_term[OF this, simplified] have "u \<cdot> \<mu> ^^ (k * M) = u" unfolding \<sigma> subst_pow_mult by auto
            hence "u \<cdot> \<mu> ^^ (M + (k - 1) * M) = u" using k by (cases k, auto)
            from this[unfolded subst_power_compose_distrib subst_compose_apply_term_distrib  cps[OF cp(1)]]
            show False unfolding u using nvar by auto
          qed
          ultimately have supt: "u \<cdot> \<sigma> \<rhd> u" by auto
          let ?u = "u \<cdot> \<sigma> \<cdot> \<mu> ^^ (M + ?n1)"
          let ?u' = "?u \<cdot> \<mu> ^^ (n3 * k)"
          from arg_cong[OF id[of k], of "\<lambda> t. t \<cdot> \<mu> ^^ n1"]  
          have "u \<cdot> \<mu> ^^ (k + M + ?n1) \<cdot> \<mu> ^^ (n3 * k) = u \<cdot> \<mu> ^^ (M + ?n1)" 
            unfolding add_commute[of k]
            unfolding
              subst_power_compose_distrib subst_compose_apply_term_distrib cps[OF cp(1)]
            unfolding 
              subst_power_compose_distrib[symmetric] subst_compose_apply_term_distrib[symmetric] by (auto simp: field_simps)
          hence "?u' = u \<cdot> \<mu> ^^ (M + ?n1)" unfolding \<sigma> subst_power_compose_distrib by simp
          with supt_stable[OF supt, of "\<mu> ^^ (M + ?n1)"]
          have "?u \<rhd> ?u'" by simp
          from supt_size[OF this] size_subst[of ?u "\<mu> ^^ (n3 * k)"] show False by auto
        qed
      qed
    }
    hence c5: "?c5 = False" by auto
    note simp = simp[unfolded c5]  
    let ?recsM = "mapM (\<lambda>(u,v,m). ident_solve' (insert (u,v,m) cps) (u \<cdot> \<mu>, v, Suc m)) (conflict_pos (s,t,N))"
    let ?res = "?recsM \<guillemotright>= (\<lambda> is. Some (max_list (map (\<lambda> i. Suc i) is)))"
    from simp
    have id: "ident_solve' cps (s,t,N) = ?res"
      by simp
    {
      fix u v m
      assume mem: "(u,v,m) \<in> ?cp"
      from conflict_pos[OF mem] obtain p' k where N: "N = m + k" and var: "is_Var v \<longrightarrow> m = 0" and eroot: "eroot u \<noteq> eroot v" and p': "p' \<in> poss s" "p' \<in> poss (t \<cdot> \<mu> ^^ k)" and uv: 
        "u = s |_ p' \<and> v = (t \<cdot> \<mu> ^^ k |_ p') \<or> v = (s |_ p') \<and> u = (t \<cdot> \<mu> ^^ k |_ p') \<and> m = 0" (is "?one \<or> ?two") by blast
      from c1 c2 c4 c5 have cs: "\<not> ?c1" "\<not> ?c2" "\<not> ?c4" "\<not> ?c5" by simp_all
      note poss = poss_imp_subst_poss[of _ _ \<mu>]
      from poss[OF p(1)] poss[OF p(2)] have pp: "p \<in> poss (S \<cdot> \<mu> ^^ Suc n)" "p \<in> poss (T \<cdot> \<mu> ^^ Suc n)"
        unfolding subst_power_Suc by simp_all
      note pp' =  poss[OF p'(1)] poss[OF p'(2)]  
      have "(S \<cdot> \<mu> ^^ Suc n |_ p, T \<cdot> \<mu> ^^ Suc n |_ p) \<in> {(s \<cdot> \<mu>, t \<cdot> \<mu> ^^ (m + k) \<cdot> \<mu>), (t \<cdot> \<mu> ^^ (m + k) \<cdot> \<mu>, s \<cdot> \<mu>)}" unfolding subst_power_Suc  
        using st unfolding N by (auto simp: subt_at_subst[OF p(1)] subt_at_subst[OF p(2)])
      hence "(S \<cdot> \<mu> ^^ Suc n |_ p, T \<cdot> \<mu> ^^ Suc n |_ p) \<in> {(s \<cdot> \<mu>, t \<cdot> \<mu> ^^ k \<cdot> \<mu> ^^ Suc m), (t \<cdot> \<mu> ^^ k \<cdot> \<mu> ^^ Suc m, s \<cdot> \<mu>)}" (is "(?S,?T) \<in> {(?s,?t),(?t,?s)}")
        unfolding add_commute[of m] subst_power_compose_distrib subst_power_Suc by simp
      hence ST: "?S = ?s \<and> ?T = ?t \<or> ?S = ?t \<and> ?T = ?s" (is "?one' \<or> ?two'") by auto
      let ?S' = "S \<cdot> \<mu> ^^ Suc n |_ p <#> p'"
      let ?T' = "T \<cdot> \<mu> ^^ Suc n |_ p <#> p'"
      have pp'S: "p <#> p' \<in> poss (S \<cdot> \<mu> ^^ (Suc n))" 
        by (rule pos_append_poss[OF pp(1)], insert ST p', auto)
      have pp'T: "p <#> p' \<in> poss (T \<cdot> \<mu> ^^ (Suc n))" 
        by (rule pos_append_poss[OF pp(2)], insert ST p', auto)
      from imageI[OF st, of "\<lambda>(s,t). (s \<cdot> \<mu> |_ p', t \<cdot> \<mu> |_ p')", unfolded split] 
      have "(?S', ?T') \<in> {(s \<cdot> \<mu> |_ p', t \<cdot> \<mu> ^^ N \<cdot> \<mu> |_ p'), (t \<cdot> \<mu> ^^ N \<cdot> \<mu> |_ p', s \<cdot> \<mu> |_ p')}" (is "_ \<in> ?l")
        unfolding image_insert split
        unfolding subt_at_subst[OF p(1), symmetric] subt_at_subst[OF p(2), symmetric] 
        unfolding subst_power_Suc[symmetric] subst_compose_apply_term_distrib[symmetric]
        unfolding subt_at_append[OF pp(1)] subt_at_append[OF pp(2)] by auto 
      moreover have "?l = {(u \<cdot> \<mu>, v \<cdot> \<mu> ^^ Suc m), (v \<cdot> \<mu> ^^ Suc m, u \<cdot> \<mu>)}"
        using uv unfolding N unfolding subt_at_subst[OF p'(1)]
        unfolding add_commute[of m] subst_power_compose_distrib subst_compose_apply_term_distrib 
        unfolding subst_compose_apply_term_distrib[of _ _ \<mu>, symmetric] subst_power_Suc[symmetric]
        unfolding subt_at_subst[OF p'(2)] by auto
      ultimately
      have S'T': "(?S',?T') \<in> {(u \<cdot> \<mu>, v \<cdot> \<mu> ^^ Suc m), (v \<cdot> \<mu> ^^ Suc m, u \<cdot> \<mu>)}" by simp
      have "\<exists> i. ident_solve' (insert (u,v,m) cps) (u \<cdot> \<mu>, v, Suc m) = Some i"
      proof (rule 1[OF refl cs mem refl refl, THEN spec, THEN spec, THEN mp, THEN mp, THEN mp, THEN mp],
        rule pp'S, rule pp'T, rule S'T', intro allI impI)
        fix u' v' m'
        assume "(u',v',m') \<in> insert (u,v,m) cps"
        thus "?Q u' v' m' (p <#> p') (Suc n)"
        proof
          assume "(u',v',m') = (u,v,m)"
          hence uv': "u' = u" "v' = v" "m' = m" by auto
          show ?thesis unfolding uv'
          proof (intro exI conjI)
            show "p <#> p' \<in> poss (S \<cdot> \<mu> ^^ n)"
              by (rule pos_append_poss[OF p(1)], insert st p', unfold N add_commute[of m] 
                subst_power_compose_distrib, auto)
          next
            show "p <#> p' \<in> poss (T \<cdot> \<mu> ^^ n)"
              by (rule pos_append_poss[OF p(2)], insert st p', unfold N add_commute[of m] 
                subst_power_compose_distrib, auto)
          next
            show "(S \<cdot> \<mu> ^^ n |_ p <#> p', T \<cdot> \<mu> ^^ n |_ p <#> p') \<in> {(u,v \<cdot> \<mu> ^^ m), (v \<cdot> \<mu> ^^ m, u)}"
              using uv st unfolding subt_at_append[OF p(1)] subt_at_append[OF p(2)] N add_commute[of m] 
              subst_power_compose_distrib subst_compose_apply_term_distrib 
              by (auto simp: subt_at_subst[OF p'(2)])
          qed (insert eroot var, auto)
        next
          assume "(u',v',m') \<in> cps"
          from CPS[THEN spec, THEN spec, THEN spec, THEN mp[OF _ this]]            
          obtain p'' n' where P: "?P u' v' m' p'' n' p n" by auto
          hence "p'' \<le> p" by simp
          also have "... \<le> p <#> p'" unfolding less_eq_pos_def by blast
          finally have "?P u' v' m' p'' n' (p <#> p') (Suc n)" using P by auto
          thus ?thesis by blast
        qed
      qed
    } note ind = this
    hence "?recsM \<noteq> None" unfolding mapM_None by auto
    thus "\<exists> i. ident_solve' cps (s,t,N) = Some i" unfolding simp by (cases ?recsM, auto)
  qed
qed

declare ident_solve'.simps[simp del]

definition ident_solve :: "('f,'v)ident_prob \<Rightarrow> nat option"
  where "ident_solve \<equiv> \<lambda> (s,t). ident_solve' {} (s,t,0)"

lemma ident_solve: 
  "ident_solve st = Some i \<Longrightarrow> ident_solution st i"
  "ident_solve st = None \<Longrightarrow> \<not> (\<exists>i. ident_solution st i)"
proof -
  assume sol: "ident_solve st = Some i"
  obtain s t where st: "st = (s,t)" by force
  from ident_solve'_sound[THEN mp[OF _ sol[unfolded st ident_solve_def split]]] show "ident_solution st i"
    unfolding ident_solution_def nident_solution_def st by auto
next
  assume non: "ident_solve st = None"
  obtain s t where st: "st = (s,t)" by force
  from non[unfolded ident_solve_def st split]
    ident_solve'_complete[of s t _ s t 0 "{}", THEN spec[of _ \<epsilon>], THEN spec[of _ 0]]
  show "\<not> (\<exists> i. ident_solution st i)" unfolding st by auto
qed

end

end

